import logging
import pathlib
import enum

import mlx.core as mx

from .llm import LLM
from .tokenizer import GGUFTokenizer, TransformerTokenizer, SentencePieceTokenizer, TiktokenTokenizer
from sillm.models.args import ModelArgs
from sillm.utils.mapping import map_key, map_config

logger = logging.getLogger("sillm")

class ModelFormat(enum.Enum):
    """
    Model type enumeration.
    """
    UNKNOWN = 0
    MLX = 1
    GGUF = 2
    HUGGINGFACE = 3

    @staticmethod
    def guess_from_weights(weights: dict):
        """
        Guess model type from weights.
        Args:
            weights: Model weights.
        Returns:
            Model type.
        """
        for k in weights:
            if k.startswith("layers."):
                return ModelFormat.MLX
            elif k.startswith("blk."):
                return ModelFormat.GGUF
            elif k.startswith("model.layers.") or k.startswith("transformer."):
                return ModelFormat.HUGGINGFACE
            
        return ModelFormat.UNKNOWN

def load(model_path: str) -> LLM:
    """
    Load model from directory.
    Args:
        model_path: Path to model directory.
    Returns:
        SiLLM model.
    """
    model_path = pathlib.Path(model_path)

    if model_path.is_dir():
        return load_model_dir(str(model_path))
    elif model_path.is_file():
        return load_model_file(str(model_path))
    else:
        raise ValueError(f"Model path {model_path} is not a file or directory")

def load_model_file(model_path: str) -> LLM:
    """
    Load model from file.
    Args:
        model_path: Path to model file.
    """
    model_path = pathlib.Path(model_path)

    if model_path.suffix == ".gguf":
        return load_gguf_file(str(model_path))
    else:
        raise ValueError(f"Unknown model file type: {model_path}")
    
def load_gguf_file(model_path: str) -> LLM:
    """
    Load model from GGUF file.
    Args:
        model_path: Path to GGUF file.
    Returns:
        SiLLM model.
    """
    logger.debug(f"Loading GGUF file {model_path}")
    gguf_weights, metadata = mx.load(model_path, return_metadata=True)

    # Map metadata to configuration
    config = map_config(metadata)
    model_args = ModelArgs.load_config(config)

    # Map weights keys
    weights = {}
    mapping = {}
    for gguf_key, value in gguf_weights.items():
        mlx_key = map_key(gguf_key)
        mapping[mlx_key] = gguf_key

        if mlx_key is None:
            logger.warn(f"Unknown key: {gguf_key}")
        else:
            weights[mlx_key] = value

    # Map quantization configuration
    gguf_file_type = metadata["general.file_type"].item()
    logger.debug(f"GGUF file type: {gguf_file_type}")
    quantization = None
    if gguf_file_type == 0 or gguf_file_type == 1:
        # No quantization
        pass
    elif gguf_file_type == 2 or gguf_file_type == 3:
        quantization = {"group_size": 32, "bits": 4}
    elif gguf_file_type == 7:
        quantization = {"group_size": 32, "bits": 8}
    else:
        logger.warn(f"Unsupported GGUF file type: {gguf_file_type}")
    model_args.quantization = quantization

    # Fix configuration
    model_args.fix_config(weights)
    model_args.log_config()

    # Load tokenizer
    tokenizer = GGUFTokenizer(metadata)
    logger.info("Loaded tokenizer from GGUF metadata")

    # Initialize model
    model = LLM(tokenizer, model_args)
    model.init_description(model_path)

    # Quantize model
    if quantization is not None:
        model.quantize(group_size=model_args.quantization["group_size"], bits=model_args.quantization["bits"], weights=weights)

    # Verify that all model weights are present
    model.verify_weights(weights)

    # Update model weights
    model.update_weights(weights, mapping=mapping)

    # Print model size
    total_params = model.get_size()
    logger.info(f"Loaded model weights with {total_params/10**9:.2f}B total parameters")

    return model

def load_model_dir(model_path: str) -> LLM:
    """
    Load model from directory.
    Args:
        model_path: Path to model directory.
    Returns:
        SiLLM model.
    """
    model_path = pathlib.Path(model_path)

    # Load configuration
    model_args = None
    for config_file in ("config.json", "params.json"):
        config_path = model_path / config_file
        if config_path.exists():
            model_args = ModelArgs.load_file(config_path)
            break
        else:
            logger.debug(f"No config file {config_path} not found")
    if model_args is None:
        raise ValueError(f"Configuration could not be loaded from {model_path}")
    logger.info(f"Loaded model config from {config_path}")

    # Load weights
    weights_files_npz = sorted(list(model_path.glob("weights*.npz")))
    weights_files_safetensors = sorted(list(model_path.glob("*.safetensors")))
    weights_files_consolidated = sorted(list(model_path.glob("consolidated.*.pth")))
    weights_files_bin = sorted(list(model_path.glob("pytorch_model-*.bin")))

    if len(weights_files_npz) > 0:
        weights, mapping, model_format = load_weights(weights_files_npz, load_func=mx.load)
    elif len(weights_files_safetensors) > 0:
        weights, mapping, model_format = load_weights(weights_files_safetensors, load_func=mx.load)
    elif len(weights_files_consolidated) > 0:
        weights, mapping, model_format = load_weights(weights_files_consolidated, load_func=load_torch_file)
    elif len(weights_files_bin) > 0:
       weights, mapping, model_format = load_weights(weights_files_bin, load_func=load_torch_file)
    else:
        raise ValueError("No weights files found")

    if model_format == ModelFormat.HUGGINGFACE:
        logger.debug("Disabling rope_traditional for HuggingFace model")

        model_args.rope_traditional = False

    # Fix configuration
    model_args.fix_config(weights)
    model_args.log_config()

    # Load tokenizer
    tokenizer = None
    tokenizer_path = None
    if (model_path / "tokenizer.model").exists():
        tokenizer_path = model_path / "tokenizer.model"
        tokenizer = SentencePieceTokenizer(str(tokenizer_path), model_args)
    elif (model_path / "tokenizer.json").exists():
        tokenizer_path = model_path / "tokenizer.json"
        tokenizer = TransformerTokenizer(str(model_path), model_args)
    elif model_args.model_type == "dbrx":
        tokenizer_path = "tiktoken"
        tokenizer = TiktokenTokenizer(model_args)
    if tokenizer is None:
        logger.error(f"No tokenizer found in {model_path}")
    logger.info(f"Loaded tokenizer from {tokenizer_path}")

    # Initialize model
    model = LLM(tokenizer, model_args)
    model.init_description(model_path)

    # Quantize model
    if model_args.quantization is not None:
        model.quantize(group_size=model_args.quantization["group_size"], bits=model_args.quantization["bits"], weights=weights)

    # Verify that all model weights are present
    model.verify_weights(weights)

    # Update model weights
    model.update_weights(weights, mapping=mapping)

    # Print model size
    total_params = model.get_size()
    logger.info(f"Loaded model weights with {total_params/10**9:.2f}B total parameters")

    return model

def load_weights(weights_files,
                 load_func: callable
                 ):
    """
    Load model weights.
    Args:
        weights_files: List of weights files.
        loader: File loader function.
    Returns:
        Model weights.
    """
    weights = {}
    mapping = {}
    format = ModelFormat.UNKNOWN
    for weights_path in weights_files:
        logger.debug(f"Loading model weights file {weights_path}")
        weights_shard = load_func(str(weights_path))

        # Guess model format according to key names
        if format == ModelFormat.UNKNOWN:
            format = ModelFormat.guess_from_weights(weights_shard)
            logger.debug(f"Guessing model format: {format}")

        for key, value in weights_shard.items():
            mlx_key = map_key(key)
            mapping[mlx_key] = key
            
            mx.eval(value)

            if mlx_key is None:
                logger.warning(f"Unknown key: {key}")
            else:
                if mlx_key in weights:
                    logger.warning(f"Duplicate key: {mlx_key} {value.shape}")

                weights[mlx_key] = value

    return weights, mapping, format

def load_torch_file(weights_path):
    """
    Load PyTorch weights and convert to MLX.
    Args:
        weights_path: Path to weights file.
    Returns:
        Model weights.
    """
    try:
        import torch
    except ImportError:
        raise ImportError("Please install torch library to load PyTorch weights")
    
    weights = {}
    for key, value in torch.load(weights_path, map_location="cpu").items():
        # Convert to numpy
        if value.dtype == torch.bfloat16:
            value = value.to(dtype=torch.float16).numpy()
        else:
            value = value.numpy()

        weights[key] = mx.array(value, dtype=mx.float16)

    return weights

########
# Based on:
# https://github.com/ggerganov/llama.cpp/blob/b7b74cef36a93ae01e0b9af8986d131761742d0e/convert.py#L1182
########
def permute_hf_weights(weights: dict,
                       args: ModelArgs
                       ):
    """
    Permute weights stored in huggingface format.
    """
    def permute(x, n):
        return (x.reshape(n, 2, x.shape[0] // n // 2, *x.shape[1:]).swapaxes(1, 2).reshape(x.shape))

    n_heads = args.n_heads
    n_kv_heads = args.n_kv_heads

    for i in range(args.n_layers):
        if f"layers.{i}.attention.wq.weight" in weights:
            weights[f"layers.{i}.attention.wq.weight"] = permute(weights[f"layers.{i}.attention.wq.weight"], n_heads)
        if f"layers.{i}.attention.wk.weight" in weights:
            weights[f"layers.{i}.attention.wk.weight"] = permute(weights[f"layers.{i}.attention.wk.weight"], n_kv_heads)


import logging
import time
import pathlib
import json

import mlx.core as mx
import mlx.nn as nn
from mlx.utils import tree_flatten, tree_unflatten

from .tokenizer import Tokenizer
import sillm.models as models
import sillm.models.args as args
from sillm.training.dataset import Dataset

logger = logging.getLogger("sillm")

model_map = {
    "llama":        models.llama.Model,
    "mistral":      models.llama.Model,
    "gemma":        models.gemma.Model,
    "mixtral":      models.mixtral.Model,   
    "phi":          models.phi.Model,
    "starcoder2":   models.starcoder2.Model,
    "qwen2":        models.qwen2.Model,
    "dbrx":         models.dbrx.Model,
    "cohere":       models.cohere.Model,
}

class LLM():
    """
    LLM model wrapper.
    """
    def __init__(self,
                 tokenizer: Tokenizer,
                 args: args.ModelArgs
                 ):
        """
        Args:
            tokenizer: Tokenizer instance.
            args: Model arguments.
        """
        self.args = args
        self._quantization = None
        self._mapping = None

        if args.model_type in model_map:
            model_class = model_map[args.model_type]
            self.model = model_class(args)
        else:
            raise NotImplementedError(f"Model type {args.model_type} is not supported")
        self.model.train(mode=False)

        self.tokenizer = tokenizer

    def init_description(self, model_path):
        """
        Set model description.
        Args:
            description: Model description.
        """
        self.path = pathlib.Path(model_path)
        self.id = self.path.name
        if self.path.is_file():
            self.id = self.path.stem    
        self.created = int(self.path.stat().st_ctime)
    
    def description(self):
        if self.id and self.created:
            return {
                "id": self.id,
                "object": "model",
                "created": self.created
            }
        else:
            raise ValueError("Model is missing an ID or creation time")

    def get_size(self):
        """
        Get model size.
        """
        total_size = 0
        for module in self.model.modules():
            if isinstance(module, nn.QuantizedLinear):
                total_size += module.weight.size * (32 // module.bits)
            elif isinstance(module, nn.Linear):
                total_size += module.weight.size

        return total_size

    def update_weights(self,
                       weights: dict,
                       mapping: dict = None
                       ):
        """
        Update model weights.
        Args:
            weights: Weights to update.
        """
        # Update weights
        weights = tree_unflatten(list(weights.items()))
        self.model.update(weights)
        mx.eval(self.model.parameters())

        # Add key mapping for potential quantization
        for key1 in list(mapping.keys()):
            key2 = mapping[key1]
            if key1.endswith(".weight"):
                mapping[key1.removesuffix(".weight") + ".scales"] = key2.removesuffix(".weight") + ".scales"
                mapping[key1.removesuffix(".weight") + ".biases"] = key2.removesuffix(".weight") + ".biases"
        
        # Update key mapping
        self._mapping = mapping

    def verify_weights(self,
                       weights: dict
                       ) -> bool:
        """
        Verify that weights for all modules are present and shapes match.
        Args:
            weights: Weights to verify.
        Returns:
            True if all weights are present and shapes match, False otherwise.
        """
        model_params = tree_flatten(self.model.parameters())
        result = True

        for name, weight in model_params:
            if name not in weights:
                result = False

                logger.warn(f"Key {name} not found in weights")
            elif weight.shape != weights[name].shape:
                result = False

                logger.warn(f"Shape mismatch for key {name}: {weight.shape} != {weights[name].shape}")

        model_keys = {name for name, _ in model_params}
        for name in weights:
            if name not in model_keys:
                logger.debug(f"Unused key {name} in weights")

        return result

    def save_weights(self,
                     weights_path: str
                     ):
        """
        Save model weights into a single safetensors file.
        Args:
            weights_path: Path to weights file.
        """
        state = dict(tree_flatten(self.model.parameters()))
        mx.save_safetensors(weights_path, state)

        logger.info(f"Saved model weights to {weights_path}")

    def save_shards(self,
                    weights_path: str,
                    max_shard_size: int = 5<<30
                    ):
        """
        Save model weights into shards.
        Args:
            weights_dir: Path to weights directory.
        """
        weights_path = pathlib.Path(weights_path)

        state = dict(tree_flatten(self.model.parameters()))

        shards = []
        weight_map = {}
        shard, shard_size = {}, 0
        total_size = 0
        for key, value in state.items():
            if self._mapping is not None:
                key = self._mapping[key]

            if shard_size + value.nbytes > max_shard_size:
                shards.append(shard)
                shard, shard_size = {}, 0

            shard[key] = value
            weight_map[key] = len(shards)

            shard_size += value.nbytes
            total_size += value.nbytes
        shards.append(shard)

        def save_shard(shard, shard_path):
            mx.save_safetensors(shard_path, shard)
            logger.debug(f"Saved shard to {shard_path}")

        if len(shards) > 1:
            for i, shard in enumerate(shards):
                save_shard(shard, str(weights_path / f"model-{i+1:05d}-of-{len(shards):05d}.safetensors"))

            for key, i in weight_map.items():
                weight_map[key] = f"model-{i+1:05d}-of-{len(shards):05d}.safetensors"
        else:
            save_shard(shard, str(weights_path / "model.safetensors"))

            for key in weight_map:
                weight_map[key] = "model.safetensors"

        index_path = str(weights_path / "model.safetensors.index.json")
        with open(index_path, "w") as f:
            index_data = {
                "metadata": {
                    "total_size": total_size,
                },
                "weight_map": weight_map
            }

            f.write(json.dumps(index_data, indent=4))
            logger.debug(f"Saved weight index to {index_path}")

    def save(self,
             model_path: str,
             max_shard_size: int = 5<<30
             ):
        """
        Save model.
        Args:
            model_path: Path to model directory.
        """
        model_path = pathlib.Path(model_path)
        model_path.mkdir(parents=True, exist_ok=True)

        # Save model config
        config_path = model_path / "config.json"
        self.args.save_config(config_path)
        logger.debug(f"Saved model config to {config_path}")

        # Save tokenizer
        self.tokenizer.save(model_path)
        logger.debug(f"Saved tokenizer to {model_path}")

        # Save model weights
        self.save_shards(model_path, max_shard_size=max_shard_size)

        logger.info(f"Saved model to {model_path}")

    def quantize(self,
                 group_size: int = 32,
                 bits: int = 4,
                 weights: dict = None,
                 ):
        """
        Quantize model.
        Args:
            group_size: Group size for quantization.
            bits: Number of bits for quantization.
            excluded: List of module names to exclude from quantization.
        """
        if self._quantization is None:
            quantization = {
                "group_size": group_size,
                "bits": bits
            }
            self._quantization = quantization
            self.args.quantization = quantization

            if weights is None:
                class_predicate = lambda p, m: (isinstance(m, (nn.Linear, nn.Embedding)) and
                                                ".gate." not in p
                                                )
            else:
                class_predicate = lambda p, m: (isinstance(m, (nn.Linear, nn.Embedding)) and
                                                f"{p}.scales" in weights
                                                )
            
            nn.quantize(self.model, class_predicate=class_predicate, **quantization)

            logger.info(f"Quantized model with group size {group_size} and {bits} bits")
        else:
            logger.warn(f"Model is already quantized with group size {group_size} and {bits} bits")

    def dequantize(self):
        """
        Dequantize model.
        """
        if self._quantization is None:
            logger.warn(f"Model is not quantized")
        else:
            layers = []
            for name, module in self.model.named_modules():
                if isinstance(module, nn.QuantizedLinear):
                    bias = "bias" in module
                    weight = mx.dequantize(module.weight, module.scales, module.biases, module.group_size, module.bits).astype(mx.float16)
                    output_dims, input_dims = weight.shape
                    linear = nn.Linear(input_dims, output_dims, bias=bias)
                    linear.weight = weight
                    if bias:
                        linear.bias = module.bias

                    layers.append((name, linear))
            
            self.model.update_modules(tree_unflatten(layers))
            self._quantization = None
            self.args.quantization = None

            logger.info(f"Dequantized model")

    def perplexity(self,
                   dataset: Dataset,
                   batch_size: int = 4
                   ):
        """
        Calculate perplexity for an input text.
        Args:
            text: Input text.
        """
        for _, batch in zip(
            range(len(dataset)),
            dataset.iterate_batches(batch_size),
        ):
            losses, _, _ = self.model.loss(*batch)
        
            yield mx.exp(mx.mean(losses)).item()

    def generate(self,
                 prompt: str,
                 temperature: float = 0.0,
                 max_tokens: int = 1024,
                 flush: int = 5
                 ):
        """
        Iterator for generating tokens.
        Args:
            prompt: Prompt to start generation.
            temperature: Sampling temperature.
            max_tokens: Max number of tokens to generate.
            flush: Flush every `flush` tokens.
        Yields:
            Tuple of generated text and metadata.
        """
        yield from generate(self.model, self.tokenizer, prompt=prompt, temperature=temperature, max_tokens=max_tokens, flush=flush)

    def completion(self,
                   prompt: str,
                   temperature: float = 0.0,
                   max_tokens: int = 1024
                   ) -> str:
        """
        Generate a completion and wait for all tokens.
        Args:
            prompt: Prompt to start generation.
            temperature: Sampling temperature.
            max_tokens: Max number of tokens to generate.
        Returns:
            Generated completion.
        """
        return ''.join([t[0] for t in generate(self.model, self.tokenizer, prompt=prompt, temperature=temperature, max_tokens=max_tokens)])

def generate(model,
             tokenizer: Tokenizer,
             prompt: str,
             temperature: float = 0.0,
             max_tokens: int = 1024,
             logprobs: bool = False,
             flush: int = 5
             ):
    start = time.perf_counter()

    # Tokenize prompt
    inputs = mx.array(tokenizer.encode(prompt))

    # Define stop tokens
    stop_tokens = tokenizer.special_ids
    
    # Initialize metadata
    timing = {
        "runtime": 0.0,
        "eval_time": 0.0,
        "tokenizer_time": time.perf_counter() - start
    }
    usage = {
        "prompt_tokens": len(inputs),
        "completion_tokens": 0,
        "total_tokens": 0
    }
    metadata = {
        "timing": timing,
        "usage": usage,
        "logprobs": [],
        "finish_reason": "length"
    }

    tokens, text, buffer, prefix = [], "", [], ""
    for (token,p) , i in zip(generate_step(model, inputs, temperature, logprobs), range(max_tokens)):
        if i == 0:
            mx.eval(token)

            timing["eval_time"] = time.perf_counter() - start

        if token.item() in stop_tokens:
            metadata["finish_reason"] = "stop"
            break

        tokens.append(token.item())

        if logprobs:
            metadata["logprobs"].append(p)

        if (len(tokens) % flush) == 0:
            mx.eval(tokens)

            text = tokenizer.decode(tokens)
            window = tokenizer.decode(buffer + tokens)
            if prefix + " " + text == window:
                prefix = text
                text = " " + text
            else:
                prefix = text
            buffer = tokens
            tokens = []

            timing["runtime"] = time.perf_counter() - start
            usage["completion_tokens"] = i+1
            usage["total_tokens"] = usage["prompt_tokens"] + usage["completion_tokens"]

            yield text, metadata

    mx.eval(tokens)
    
    text = tokenizer.decode(tokens)
    window = tokenizer.decode(buffer + tokens)
    if prefix + " " + text == window:
        prefix = text
        text = " " + text
    else:
        prefix = text
    buffer = tokens
    tokens = []

    timing["runtime"] = time.perf_counter() - start
    usage["completion_tokens"] = i+1
    usage["total_tokens"] = usage["prompt_tokens"] + usage["completion_tokens"]

    yield text, metadata

def generate_step(model, inputs, temperature, logprobs=False):
    y = inputs
    cache = None
    while True:
        logits, cache = model(y[None], cache=cache)
        logits = logits[:, -1, :]
        y = sample(logits, temperature)

        p = 0.0
        if logprobs:
            p = nn.log_softmax(logits, axis=-1)[0,y].item()
        
        yield y, p

def sample(logits, temperature):
    if temperature > 0:
        return mx.random.categorical(logits * (1 / temperature))
    else:
        return mx.argmax(logits, axis=-1)
    # TODO add top-p sampling


import argparse
import pathlib

import sillm.utils as utils

from sillm.utils.quantization import quantize_files
from sillm.models.args import ModelArgs
from sillm.core.tokenizer import TransformerTokenizer, SentencePieceTokenizer

if __name__ == "__main__":
    # Parse commandline arguments
    parser = argparse.ArgumentParser(description="A simple CLI for generating text with SiLLM.")
    parser.add_argument("input", type=str, help="The input model directory or file")
    parser.add_argument("output", type=str, help="The output model directory or file")
    parser.add_argument("--bits", default=4, help="Quantization bits")
    parser.add_argument("--group_size", default=32, help="Quantization group size")
    parser.add_argument("-v", "--verbose", default=1, action="count", help="Increase output verbosity")
    args = parser.parse_args()

    # Initialize logging
    log_level = 40 - (10 * args.verbose) if args.verbose > 0 else 0
    logger = utils.init_logger(log_level)

    # Log commandline arguments
    if log_level <= 10:
        utils.log_arguments(args.__dict__)

    input_path = pathlib.Path(args.input)
    output_path = pathlib.Path(args.output)
    quantization = {
        "bits": args.bits,
        "group_size": args.group_size
    }

    quantize_files(args.input, args.output, **quantization)

    config_path = input_path / "config.json"
    model_args = ModelArgs.load_file(config_path)
    model_args.quantization = quantization
    config_path = output_path / "config.json"
    model_args.save_config(config_path)
    logger.debug(f"Saved model config to {config_path}")

    tokenizer = None
    tokenizer_path = None
    if (input_path / "tokenizer.model").exists():
        tokenizer_path = input_path / "tokenizer.model"
        tokenizer = SentencePieceTokenizer(str(tokenizer_path), model_args)
    elif (input_path / "tokenizer.json").exists():
        tokenizer_path = input_path / "tokenizer.json"
        tokenizer = TransformerTokenizer(str(input_path), model_args)
    if tokenizer is None:
        logger.error(f"No tokenizer found in {input_path}")
    tokenizer.save(str(output_path))
    logger.debug(f"Saved tokenizer to {output_path}")

    logger.info(f"Quantized model with group size {args.group_size} and {args.bits} bits")


from typing import Optional, Tuple

import mlx.core as mx
import mlx.nn as nn

from sillm.models.base import BaseModel
from sillm.models.args import ModelArgs
import sillm.models.llama as llama

class Attention(llama.Attention):
    """
    Multi-head attention module.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        nn.Module.__init__(self)
        self.args = args

        self.n_heads: int = args.n_heads
        self.n_kv_heads: int = args.n_kv_heads

        self.repeats = self.n_heads // self.n_kv_heads
        self.scale = self.args.head_dim ** -0.5
        self.rope_dims = int(args.partial_rotary_factor * args.head_dim)

        self.wq = nn.Linear(args.dim, args.n_heads * args.head_dim, bias=True)
        self.wk = nn.Linear(args.dim, args.n_kv_heads * args.head_dim, bias=True)
        self.wv = nn.Linear(args.dim, args.n_kv_heads * args.head_dim, bias=True)
        self.wo = nn.Linear(args.n_heads * args.head_dim, args.dim, bias=True)
        
        self.rope = nn.RoPE(self.rope_dims,
                            traditional=False,
                            base=args.rope_theta)
        
    def __call__(self,
                 x: mx.array,
                 mask: Optional[mx.array] = None,
                 cache: Optional[Tuple[mx.array, mx.array]] = None,
                 ) -> mx.array:
        B, L, _ = x.shape

        queries, keys, values = self.wq(x), self.wk(x), self.wv(x)

        # Prepare the queries, keys and values for the attention computation
        queries = queries.reshape(B, L, self.n_heads, -1).transpose(0, 2, 1, 3)
        keys = keys.reshape(B, L, self.n_kv_heads, -1).transpose(0, 2, 1, 3)
        values = values.reshape(B, L, self.n_kv_heads, -1).transpose(0, 2, 1, 3)

        if self.repeats > 1:
            keys = mx.repeat(keys, self.repeats, axis=1)
            values = mx.repeat(values, self.repeats, axis=1)

        if cache is not None:
            key_cache, value_cache = cache
            queries = self.rope(queries, offset=key_cache.shape[2])
            keys = self.rope(keys, offset=key_cache.shape[2])
            keys = mx.concatenate([key_cache, keys], axis=2)
            values = mx.concatenate([value_cache, values], axis=2)
        else:
            queries = self.rope(queries)
            keys = self.rope(keys)

        queries = queries.astype(mx.float32)
        keys = keys.astype(mx.float32)

        scores = (queries * self.scale) @ keys.transpose(0, 1, 3, 2)
        if mask is not None:
            scores += mask
        scores = mx.softmax(scores.astype(mx.float32), axis=-1).astype(scores.dtype)
        output = (scores @ values).transpose(0, 2, 1, 3).reshape(B, L, -1)

        return self.wo(output), (keys, values)

class FeedForward(nn.Module):
    """
    Feed-forward module.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        super().__init__()

        self.w1 = nn.Linear(args.dim, args.hidden_dim, bias=True)
        self.w2 = nn.Linear(args.hidden_dim, args.dim, bias=True)
    
    def __call__(self,
                 x: mx.array
                 ) -> mx.array:
        """
        Args:
            x: Input tensor.
        Returns:
            Output tensor.
        """
        return self.w2(nn.gelu_approx(self.w1(x)))
    
class TransformerBlock(nn.Module):
    """
    Transformer block.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        nn.Module.__init__(self)
        self.args = args

        self.n_heads = args.n_heads
        self.dim = args.dim

        self.attention = Attention(args=args)
        self.attention_norm = nn.LayerNorm(args.dim, eps=args.norm_eps)
        self.feed_forward = FeedForward(args=args)

    def forward(self,
                x: mx.array,
                mask: Optional[mx.array] = None,
                cache: Optional[Tuple[mx.array, mx.array]] = None,
                ) -> mx.array:
        h = self.attention_norm(x)
        attn_h, cache = self.attention(h, mask, cache)
        ff_h = self.feed_forward(h)

        return attn_h + ff_h + x, cache
    
class Model(llama.Model):
    """
    Phi model wrapper.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        BaseModel.__init__(self, args)
        self.args = args

        self.n_layers = args.n_layers
        self.vocab_size = args.vocab_size
        
        self.tok_embeddings = nn.Embedding(args.vocab_size, args.dim)
        self.layers = [TransformerBlock(args=args) for _ in range(args.n_layers)]
        self.norm = nn.LayerNorm(args.dim, eps=args.norm_eps)
        self.output = nn.Linear(args.dim, args.vocab_size, bias=True)


import pathlib
import shutil

from typing import List

import sillm.models.args as args

class Tokenizer():
    def encode(self,
               s: str,
               bos: bool = True,
               eos: bool = False
               ) -> List[int]:
        raise NotImplementedError("Class tokenizer.Tokenizer is used for inheritance only")
    
    def decode(self,
               t: List[int]
               ) -> str:
        raise NotImplementedError("Class tokenizer.Tokenizer is used for inheritance only")
    
    def has_template(self) -> bool:
        return False
    
    def apply_chat_template(self,
                            *args,
                            **kwargs
                            ):
        raise NotImplementedError("Tokenizer does not support chat templates - check has_template() first")
        
    @property
    def vocab_size(self) -> int:
        raise NotImplementedError("Class tokenizer.Tokenizer is used for inheritance only")
    
    @property
    def special_tokens_map(self) -> dict:
        raise NotImplementedError("Class tokenizer.Tokenizer is used for inheritance only")
    
    def save(self,
             tokenizer_path: str
             ):
        raise NotImplementedError("Class tokenizer.Tokenizer is used for inheritance only")
    
########
# Based on mlx-examples:
# https://github.com/ml-explore/mlx-examples/blob/e74889d0fa0fb49d95bfdf6a1dcad907713eb50e/llms/mixtral/mixtral.py#L221
########
class SentencePieceTokenizer(Tokenizer):
    """
    Tokenizer wrapper for SentencePiece.
    """
    def __init__(self,
                 tokenizer_path: str,
                 args: args.ModelArgs
                 ):
        """
        Args:
            tokenizer_path: Path to tokenizer file.
        """
        assert pathlib.Path(tokenizer_path).exists(), tokenizer_path
        self._source = tokenizer_path

        try:
            import sentencepiece
        except ImportError:
            raise ImportError("Please install sentencepiece library to use SentencePieceTokenizer")

        self._model = sentencepiece.SentencePieceProcessor(model_file=tokenizer_path)
        self._sep = "▁"

        assert self._model.vocab_size() == self._model.get_piece_size()

        if args.bos_token_id is None:
            self.bos_id = self._model.bos_id()
        else:
            self.bos_id = args.bos_token_id
        if args.eos_token_id is None:
            self.eos_id = self._model.eos_id()
        else:
            self.eos_id = args.eos_token_id
        self.pad_id = self._model.pad_id()

        self.special_ids = [self.bos_id, self.eos_id, self.pad_id]

    def encode(self,
               s: str,
               bos: bool = True,
               eos: bool = False
               ) -> List[int]:
        """
        Encode string.
        Args:
            s: String to encode.
            bos: Whether to prefix BOS token.
            eos: Whether to append EOS token.
        Returns:
            Encoded tokens.
        """
        if bos:
            tokens = [self.bos_id, *self._model.encode(s)]
        else:
            tokens = self._model.encode(s)

        if eos:
            tokens.append(self.eos_id)

        return tokens
    
    def decode(self,
               t: List[int]
               ) -> str:
        """
        Decode tokens.
        Args:
            t: Tokens to decode.
        Returns:
            Decoded string.
        """
        s = self._model.decode(t)
        if t and self._model.id_to_piece(t[0])[0] == self._sep:
            return " " + s
        
        return s

    @property
    def vocab_size(self) -> int:
        """
        Vocabulary size.
        """
        return self._model.vocab_size()
    
    @property
    def special_tokens_map(self) -> dict:
        """
        Special tokens map.
        """
        special_tokens_map = {
            "bos_token": self._model.id_to_piece(self.bos_id),
            "eos_token": self._model.id_to_piece(self.eos_id),
        }
        
        return special_tokens_map
    
    def save(self,
            tokenizer_path: str
            ):
        tokenizer_path = pathlib.Path(tokenizer_path) / "tokenizer.model"

        shutil.copy(self._source, tokenizer_path)
    
class TransformerTokenizer(Tokenizer):
    """
    Tokenizer wrapper for Transformers AutoTokenizer.
    """
    def __init__(self,
                 tokenizer_dir: str,
                 args: args.ModelArgs
                 ):
        """
        Args:
            tokenizer_path: Path to tokenizer directory.
        """
        assert pathlib.Path(tokenizer_dir).exists(), tokenizer_dir

        try:
            import transformers
            transformers.logging.set_verbosity_error()
        except ImportError:
            raise ImportError("Please install transformers library to use TransformerTokenizer")

        self._model = transformers.AutoTokenizer.from_pretrained(tokenizer_dir)

        if args.bos_token_id is None:
            self.bos_id = self._model.bos_token_id
        else:
            self.bos_id = args.bos_token_id
        if args.eos_token_id is None:
            self.eos_id = self._model.eos_token_id
        else:
            self.eos_id = args.eos_token_id
        if args.pad_token_id is None:
            self.pad_id = self._model.pad_token_id
        else:
            self.pad_id = args.pad_token_id

        self.special_ids = set([self.bos_id, self.eos_id] + self._model.all_special_ids)

        # Hack for Llama-3
        if "<|eot_id|>" in self._model.vocab:
            self.special_ids.add(self._model.vocab["<|eot_id|>"])

    def encode(self,
               s: str,
               bos: bool = True,
               eos: bool = False
               ) -> List[int]:
        """
        Encode string.
        Args:
            s: String to encode.
            bos: Whether to prefix BOS token.
            eos: Whether to append EOS token.
        Returns:
            Encoded tokens.
        """
        if bos:
            tokens = [self.bos_id, *self._model.encode(s, add_special_tokens=False)]
        else:
            tokens = self._model.encode(s, add_special_tokens=False)

        if eos:
            tokens.append(self.eos_id)

        return tokens
    
    def decode(self,
               t: List[int]
               ) -> str:
        """
        Decode tokens.
        Args:
            t: Tokens to decode.
        Returns:
            Decoded string.
        """
        return self._model.decode(t)
    
    def has_chat_template(self) -> bool:
        """
        Check if tokenizer has chat template.
        """
        return hasattr(self._model, "apply_chat_template")
    
    def apply_chat_template(self,
                            *args,
                            **kwargs
                            ):
        """
        Apply chat template.
        """
        return self._model.apply_chat_template(*args, **kwargs)

    @property
    def vocab_size(self) -> int:
        """
        Vocabulary size.
        """
        return self._model.vocab_size
    
    @property
    def special_tokens_map(self) -> dict:
        """
        Special tokens map.
        """
        return self._model.special_tokens_map
    
    def save(self,
            tokenizer_path: str
            ):
        self._model.save_pretrained(tokenizer_path)

########
# Based on mlx-examples:
# https://github.com/ml-explore/mlx-examples/blob/854ad8747a9c703773adf8866602b114f68aa54a/llms/gguf_llm/utils.py#L5
########
class GGUFTokenizer(SentencePieceTokenizer):
    """
    Tokenizer wrapper for GGUF.
    """
    def __init__(self,
                 metadata: dict
                 ):
        """
        Args:
            metadata: Model metadata in GGUF format.
        """
        try:
            import sentencepiece
            import sentencepiece.sentencepiece_model_pb2
        except ImportError:
            raise ImportError("Please install sentencepiece and protobuf==3.20.1 libraries to use SentencePieceTokenizer")
        
        tokens = metadata["tokenizer.ggml.tokens"]
        scores = metadata.get("tokenizer.ggml.scores", None)
        scores = scores.tolist() if scores is not None else None
        token_types = metadata.get("tokenizer.ggml.token_type", None)
        byte_fallback = False

        self.bos_id = metadata["tokenizer.ggml.bos_token_id"].item()
        self.eos_id = metadata["tokenizer.ggml.eos_token_id"].item()
        self.unk_id = -1
        if "tokenizer.ggml.unknown_token_id" in metadata:
            self.unk_id = metadata["tokenizer.ggml.unknown_token_id"].item()
        if "tokenizer.ggml.padding_token_id" in metadata:
            self.pad_id = metadata["tokenizer.ggml.padding_token_id"].item()
            pad_token = tokens[self.pad_id]
        else:
            self.pad_id = -1
            pad_token = "<pad>"
        self._sep = "▁"

        self.special_ids = set([self.bos_id, self.eos_id])
        
        if token_types is not None:
            token_types = token_types.tolist()

            # Determine if there enough tokens to use byte fallback
            if token_types.count(6) >= 256:
                # 6 is the token type for BYTE
                byte_fallback = True
            
            # Find UNK token if not provided in metadata
            if self.unk_id < 0:
                for i in range(len(token_types)):
                    if token_types[i] == 2:
                        self.unk_id = i

        # Add UNK token if not defined
        if self.unk_id < 0:
            self.unk_id = len(tokens)
            tokens.append("<unk>")
            if scores:
                scores.append(0)
            if token_types:
                # 2 is the token type for UNKNOWN
                token_types.append(2)

        trainer_spec = sentencepiece.sentencepiece_model_pb2.TrainerSpec(
            model_type="BPE",
            vocab_size=len(tokens),
            input_format="text",
            split_by_unicode_script=True,
            split_by_whitespace=True,
            split_by_number=True,
            treat_whitespace_as_suffix=False,
            split_digits=True,
            allow_whitespace_only_pieces=True,
            vocabulary_output_piece_score=True,
            byte_fallback=byte_fallback,
            unk_id=self.unk_id,
            bos_id=self.bos_id,
            eos_id=self.eos_id,
            pad_id=self.pad_id,
            unk_piece=tokens[self.unk_id],
            bos_piece=tokens[self.bos_id],
            eos_piece=tokens[self.eos_id],
            pad_piece=pad_token,
            pretokenization_delimiter="",
        )
        normalizer_spec = sentencepiece.sentencepiece_model_pb2.NormalizerSpec(
            name="identity",
            precompiled_charsmap=b"",
            add_dummy_prefix=True,
            remove_extra_whitespaces=False,
            normalization_rule_tsv=b"",
        )
        model_proto = sentencepiece.sentencepiece_model_pb2.ModelProto(trainer_spec=trainer_spec, normalizer_spec=normalizer_spec)

        for i, token in enumerate(tokens):
            score = scores[i] if scores else 0
            token_type = token_types[i] if token_types else 0
            model_proto.pieces.append(sentencepiece.sentencepiece_model_pb2.ModelProto.SentencePiece(piece=token, score=score, type=token_type))
        
        self._model = sentencepiece.SentencePieceProcessor(model_proto=model_proto.SerializeToString())

    def save(self,
            tokenizer_path: str
            ):
        raise NotImplementedError("GGUFTokenizer does not support saving")

class TiktokenTokenizer(Tokenizer):
    """
    Tokenizer wrapper for tiktoken.
    """
    def __init__(self,
                 args: args.ModelArgs,
                 model_name: str = "gpt-4"
                 ):
        try:
            import tiktoken
        except ImportError:
            raise ImportError("Please install tiktoken library to use TiktokenTokenizer")
        
        self._model = tiktoken.encoding_for_model(model_name)

        self.bos_id = args.bos_token_id
        self.eos_id = args.eos_token_id

        self.special_ids = set([self.bos_id, self.eos_id])

    def encode(self,
               s: str,
               bos: bool = True,
               eos: bool = False
               ) -> List[int]:
        """
        Encode string.
        Args:
            s: String to encode.
            bos: Whether to prefix BOS token.
            eos: Whether to append EOS token.
        Returns:
            Encoded tokens.
        """
        tokens = self._model.encode(s)
        if bos:
            tokens = [self.bos_id] + tokens
        if eos:
            tokens.append(self.eos_id)

        return tokens
    
    def decode(self,
               t: List[int]
               ) -> str:
        """
        Decode tokens.
        Args:
            t: Tokens to decode.
        Returns:
            Decoded string.
        """
        return self._model.decode(t)
    
    @property
    def vocab_size(self) -> int:
        """
        Vocabulary size.
        """
        return self._model.n_vocab
    
    def save(self,
            tokenizer_path: str
            ):
        raise NotImplementedError("TiktokenTokenizer does not support saving")


from typing import Optional, Tuple

import numpy as np

import mlx.core as mx
import mlx.nn as nn

from sillm.models.base import BaseModel
from sillm.models.args import ModelArgs
import sillm.models.llama as llama

########
# Based on mlx-examples:
# https://github.com/ml-explore/mlx-examples/blob/1415595409874971b5ad96d55980e7d4aa8a8043/lora/models/mixtral.py#L141
########
class FeedForward(nn.Module):
    """
    MoE Feed-forward module for Mixtral models.
    """
    def __init__(self, args: ModelArgs):
        super().__init__()

        self.num_experts = args.moe["num_experts"]
        self.num_experts_per_tok = args.moe["num_experts_per_tok"]

        self.gate = nn.Linear(args.dim, self.num_experts, bias=False)
        self.experts = [llama.FeedForward(args=args) for _ in range(self.num_experts)]

    def __call__(self,
                 x: mx.array,
                 training_loss: bool = False
                 ) -> mx.array:
        """
        Args:
            x: Input tensor.
        Returns:
            Output tensor.
        """
        top_k = self.num_experts_per_tok
        orig_shape = x.shape
        x = x.reshape(-1, x.shape[-1])

        gate_logits = self.gate(x)
        
        expert_indices = mx.stop_gradient(mx.argpartition(-gate_logits, kth=top_k-1, axis=-1)[:, :top_k])
        expert_scores = mx.softmax(mx.take_along_axis(gate_logits, expert_indices, axis=-1).astype(mx.float32), axis=-1).astype(gate_logits.dtype)
        
        if self.training:
            y = mx.zeros((x.shape[0], self.num_experts_per_tok, x.shape[-1]))
            for e, expert in enumerate(self.experts):
                idx1, idx2 = map(mx.array, np.where(expert_indices == e))
                if idx1.size == 0:
                    continue
                y[idx1, idx2] = expert(x[idx1])

            y = (y * expert_scores[:, :, None]).sum(axis=1)
        else:
            y = []
            for xt, st, it in zip(x, expert_scores, expert_indices.tolist()):
                yt = mx.concatenate([self.experts[e](xt)[:, None] for e in it], axis=-1)
                yt = (yt * st).sum(axis=-1)
                y.append(yt[None, :])
            y = mx.concatenate(y)

        if training_loss:
            ########
            # Calculate expert router loss
            # References:
            # https://github.com/huggingface/transformers/blob/19e83d174c1e2802a459c9b5831628817e1c286f/src/transformers/models/mixtral/modeling_mixtral.py#L77
            # https://arxiv.org/abs/2101.03961 (Switch Transformers: Scaling to Trillion Parameter Models with Simple and Efficient Sparsity)
            ########
            
            # Calculate routing weights and router probability per expert
            routing_weights = mx.softmax(gate_logits, axis=-1)
            router_prob_per_expert = mx.mean(routing_weights, axis=0)
            
            # Calculate expert counts and tokens per expert
            expert_counts = np.bincount(expert_indices.flatten(), minlength=8)
            tokens_per_expert = expert_counts / expert_counts.sum()
            
            # Calculate expert loss and total router loss
            expert_losses = tokens_per_expert * router_prob_per_expert
            router_loss = expert_losses.sum()
        else:
            router_loss = 0.0

        return y.reshape(orig_shape), router_loss
    
########
# Based on mlx-examples:
# https://github.com/ml-explore/mlx-examples/blob/d8680a89f986492dbc27c36af3294034db26458f/llms/mixtral/mixtral.py#L163
########
class TransformerBlock(nn.Module):
    """
    Transformer block for Mixtral models.
    """
    def __init__(self, args: ModelArgs):
        super().__init__()

        self.n_heads = args.n_heads
        self.dim = args.dim
        self.attention = llama.Attention(args=args)
        self.feed_forward = FeedForward(args=args)
        self.attention_norm = nn.RMSNorm(args.dim, eps=args.norm_eps)
        self.ffn_norm = nn.RMSNorm(args.dim, eps=args.norm_eps)
        self.args = args

    def forward(self,
                 x: mx.array,
                 mask: mx.array = None,
                 cache = None,
                 training_loss: bool = False
                 ) -> mx.array:
        r, cache = self.attention(self.attention_norm(x), mask, cache)
        h = x + r
        r, router_loss = self.feed_forward(self.ffn_norm(h), training_loss)
        out = h + r

        return out, cache, router_loss
    
########
# Based on mlx-examples:
# https://github.com/ml-explore/mlx-examples/blob/e74889d0fa0fb49d95bfdf6a1dcad907713eb50e/llms/mixtral/mixtral.py#L187
########
class Model(BaseModel):
    """
    Mixtral model wrapper.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        super().__init__(args)
        self.args = args

        self.n_layers = args.n_layers
        self.vocab_size = args.vocab_size
        self.router_aux_loss_coef = args.router_aux_loss_coef

        self.tok_embeddings = nn.Embedding(args.vocab_size, args.dim)
        self.layers = [TransformerBlock(args=args) for _ in range(args.n_layers)]
        self.norm = nn.RMSNorm(args.dim, eps=args.norm_eps)
        self.output = nn.Linear(args.dim, args.vocab_size, bias=False)

    def __call__(self,
                 inputs: mx.array,
                 cache = None
                 ):
        """
        Args:
            inputs: Input tokens.
            cache: Cache from previous forward pass.
        Returns:
            Output logits and cache.
        """
        h = self.tok_embeddings(inputs)

        mask = None
        T = h.shape[1]
        if T > 1:
            mask = nn.MultiHeadAttention.create_additive_causal_mask(T)
            mask = mask.astype(h.dtype)

        if cache is None:
            cache = [None] * len(self.layers)

        for e, layer in enumerate(self.layers):
            h, cache[e], _ = layer.forward(h, mask, cache[e])

        return self.output(self.norm(h)), cache

    def loss(self,
             inputs: mx.array,
             targets: mx.array,
             loss_masks: mx.array
             ):
        """
        Calculate loss for inputs.
        Args:
            inputs: Input tokens.
            targets: Target tokens.
            lengths: Lengths of inputs.
        Returns:
            Cross-entropy + router loss.
        """
        h = self.tok_embeddings(inputs)

        mask = None
        T = h.shape[1]
        if T > 1:
            mask = nn.MultiHeadAttention.create_additive_causal_mask(T)
            mask = mask.astype(h.dtype)

        aux_loss = 0.0
        for layer in self.layers:
            h, _, router_loss = layer(h, mask, None, training_loss=True)
            aux_loss += router_loss

        logits = self.output(self.norm(h))
        logits = logits.astype(mx.float32)

        # Calculate the loss
        cross_entropy_loss = nn.losses.cross_entropy(logits, targets) * loss_masks
        num_tokens = loss_masks.sum()
        loss_value = cross_entropy_loss.sum() / num_tokens

        overall_loss = loss_value + aux_loss * self.router_aux_loss_coef

        return overall_loss, None, num_tokens


import logging

import jinja2
import jinja2.exceptions

logger = logging.getLogger("sillm")

default_templates = {
    "llama": "llama2",
    "mistral": "mistral",
    "gemma": "gemma",
    "mixtral": "mistral",
    "phi": "phi2",
    "qwen2": "qwen2",
    "dbrx": "chatml",
    "cohere": "cohere",
}

class Template(object):
    def __init__(self,
                 tokenizer,
                 template_name: str = "chatml",
                 exception_callback = None
                 ):
        loader = jinja2.PackageLoader('sillm', 'templates')
        env = jinja2.Environment(loader=loader, trim_blocks=True, lstrip_blocks=True)

        if exception_callback is None:
            env.globals["raise_exception"] = self._raise_exception
        else:
            env.globals["raise_exception"] = exception_callback

        fname_template = template_name + ".jinja"
        self.template = env.get_template(fname_template)

        # Set special tokens map
        self.special_tokens_map = tokenizer.special_tokens_map

        logger.info(f"Initialized conversation template: {template_name}")

    @staticmethod
    def list_templates():
        loader = jinja2.PackageLoader('sillm', 'templates')

        return loader.list_templates()
    
    @staticmethod
    def guess_template(args):
        if args.model_type and args.model_type in default_templates:
            template_name = default_templates[args.model_type]

            # Hack for Llama-3
            if args.model_type == "llama" and args.vocab_size == 128256:
                template_name = "llama3"

            return template_name

        return None

    def _raise_exception(self, msg):
        raise jinja2.exceptions.TemplateError(msg)

    def apply_chat_template(self,
                            messages: list,
                            add_generation_prompt: bool = False
                            ):
        return self.template.render(messages=messages, add_generation_prompt=add_generation_prompt, **self.special_tokens_map)
    
class AutoTemplate(Template):
    def __init__(self,
                 tokenizer
                 ):
        self.tokenizer = tokenizer

        logger.info(f"Initialized built-in conversation template from tokenizer")
    
    def apply_chat_template(self,
                            messages: list,
                            add_generation_prompt: bool = False
                            ):
        return self.tokenizer.apply_chat_template(messages, tokenize=False, add_generation_prompt=add_generation_prompt)

def init_template(tokenizer,
                  args,
                  template_name: str = None
                  ):
    if template_name:
        template = Template(tokenizer, template_name=template_name)
    else:
        template_name = Template.guess_template(args)
        if template_name:
            template = Template(tokenizer, template_name=template_name)
        elif tokenizer.has_template:
            template = AutoTemplate(tokenizer)
        else:
            template = Template(tokenizer, template_name="empty")

            logger.warn("No conversation template found - falling back to empty template.")

    return template


from functools import partial

import mlx.core as mx
import mlx.nn as nn

from sillm.models.args import ModelArgs

class BaseModel(nn.Module):
    """
    Base class for LLM models.
    """
    def __init__(self, args: ModelArgs):
        super().__init__()
    
    def __call__(self,
                 inputs: mx.array,
                 cache = None
                 ):
        raise NotImplementedError("Class model.Model is used for inheritance only")
    
    def loss(self,
        inputs: mx.array,
        targets: mx.array,
        lengths: mx.array):
        raise NotImplementedError("Loss function is not implemented for this model")
    
    ########
    # Based on mlx-examples:
    # https://github.com/ml-explore/mlx-examples/blob/047d4650c4f63d55e5bfbaf8f589c1679cbdd971/lora/lora.py#L151
    ########
    def loss(self,
            inputs: mx.array,
            targets: mx.array,
            loss_masks: mx.array
            ):
        """
        Calculate cross-entropy loss.
        Args:
            inputs: Input tokens.
            targets: Target tokens.
            lengths: Lengths of inputs.
        Returns:
            Cross-entropy loss.
        """
        # Run model on inputs
        logits, _ = self.__call__(inputs)
        logits = logits.astype(mx.float32)

        # Calculate the loss
        cross_entropy_loss = nn.losses.cross_entropy(logits, targets) * loss_masks
        num_tokens = loss_masks.sum()
        loss_value = cross_entropy_loss.sum() / num_tokens

        return loss_value, None, num_tokens


import logging

import mlx.core as mx
import mlx.nn as nn

from sillm.core.llm import LLM, generate
from sillm.models.args import ModelArgs
from sillm.training.lora import TrainableLoRA

logger = logging.getLogger("sillm")

class TrainableDPO(TrainableLoRA):
    """
    Trainable DPO model wrapper.
    """
    @staticmethod
    def from_model(llm: LLM, **kwargs):
        """
        Convert LLM to trainable DPO LLM.
        Args:
            llm: LLM to convert.
        Returns:
            Trainable LLM.
        """
        model = TrainableDPO(llm.model, llm.tokenizer, llm.args, **kwargs)
        model._quantization = llm._quantization

        return model
    
    def __init__(self,
                 model,
                 tokenizer,
                 args: ModelArgs,
                 reference_free: bool = False,
                 loss_type: str = "sigmoid",
                 loss_beta: float = 0.1,
                 label_smoothing: float = 0.0
                 ):
        """
        Args:
            tokenizer: Tokenizer instance.
            args: Model arguments.
            loss_type: Type of loss function (sigmoid/hinge/ipo/dpop).
            loss_beta: Loss beta parameter.
        """
        super().__init__(model, tokenizer, args)

        self.reference_free = reference_free
        self.loss_type = loss_type
        self.beta = loss_beta
        self.label_smoothing = label_smoothing

        if reference_free:
            self.reference = None
        else:
            model_class = model.__class__
            self.reference = model_class(args)
            weights = self.model.parameters()
            self.reference.update(weights)
            self.reference.freeze()
            self.reference.train(mode=False)

        logger.info(f"Initialized DPO with reference model and loss type {loss_type}")

    def comparison(self,
                   prompt: str,
                   temperature: float = 0.0,
                   max_tokens: int = 1024
                   ):
        """
        Generate comparison between policy and reference model completions.
        Args:
            prompt: Prompt to start generation.
            temperature: Sampling temperature.
            max_tokens: Max number of tokens to generate.
        Returns:
            Completions.
        """
        reference_completion = ''.join([t[0] for t in generate(self.reference, self.tokenizer, prompt, temperature==temperature, max_tokens=max_tokens)])
        policy_completion = ''.join([t[0] for t in generate(self.model, self.tokenizer, prompt, temperature=temperature, max_tokens=max_tokens)])

        return reference_completion, policy_completion

    ########
    # References:
    # https://github.com/eric-mitchell/direct-preference-optimization
    # https://huggingface.co/docs/trl/main/en/dpo_trainer
    ########
    def loss(self,
            chosen: mx.array,
            rejected: mx.array,
            chosen_masks: mx.array,
            rejected_masks: mx.array,
            ):
        """
        Calculate loss for inputs.
        Args:
            inputs: Input tokens.
            targets: Target tokens.
            lengths: Lengths of inputs.
        Returns:
            Loss value.
        """
        def forward(model, x, mask):
            inputs = x[:, :-1]
            targets = x[:, 1:]
            
            logits, _ = model(inputs)
            logits = logits.astype(mx.float32)

            return -nn.losses.cross_entropy(logits, targets) * mask

        num_chosen_tokens = chosen_masks.sum(-1)
        num_rejected_tokens = rejected_masks.sum(-1)

        # Calculate log probabilities for policy model
        policy_chosen_scores = forward(self.model, chosen, chosen_masks)
        policy_rejected_scores = forward(self.model, rejected, rejected_masks)
        if self.loss_type == "ipo":
            # ipo uses average log probabilities
            policy_chosen_score = policy_chosen_scores.sum(-1) / num_chosen_tokens
            policy_rejected_score = policy_rejected_scores.sum(-1) / num_rejected_tokens
        else:
            policy_chosen_score = policy_chosen_scores.sum(-1)
            policy_rejected_score = policy_rejected_scores.sum(-1)

        # Calculate log probabilities for reference model
        if self.reference_free:
            reference_chosen_score = mx.zeros_like(policy_chosen_score)
            reference_rejected_score = mx.zeros_like(policy_rejected_score)
        else:
            reference_chosen_scores = mx.stop_gradient(forward(self.reference, chosen, chosen_masks))
            reference_rejected_scores = mx.stop_gradient(forward(self.reference, rejected, rejected_masks))
            if self.loss_type == "ipo":
                # ipo uses average log probabilities
                reference_chosen_score = reference_chosen_scores.sum(-1) / num_chosen_tokens
                reference_rejected_score = reference_rejected_scores.sum(-1) / num_rejected_tokens
            else:
                reference_chosen_score = reference_chosen_scores.sum(-1)
                reference_rejected_score = reference_rejected_scores.sum(-1)
        
        logits = (policy_chosen_score - policy_rejected_score) - (reference_chosen_score - reference_rejected_score)

        if self.loss_type == "sigmoid":
            # https://arxiv.org/abs/2305.18290
            losses = -nn.log_sigmoid(self.beta * logits)
        elif self.loss_type == "hinge":
            # https://arxiv.org/abs/2309.06657
            losses = nn.relu(1 - self.beta * logits)
        elif self.loss_type == "ipo":
            # https://arxiv.org/abs/2310.12036
            losses = (logits - 1 / (2 * self.beta)) ** 2
        elif self.loss_type == "dpop":
            # https://arxiv.org/abs/2402.13228v1
            self.delta = 50
            penalty = mx.maximum(mx.zeros_like(policy_chosen_score), reference_chosen_score - policy_chosen_score)
            losses = -(nn.log_sigmoid(self.beta * logits) - self.delta * penalty)
        else:
            raise ValueError(f"Unknown loss type: {self.loss_type}")
        
        loss = mx.mean(losses)
        num_tokens = (num_chosen_tokens + num_rejected_tokens).sum()

        chosen_reward = self.beta * mx.mean(policy_chosen_score - reference_chosen_score)
        rejected_reward = self.beta * mx.mean(policy_rejected_score - reference_rejected_score)
        reward = mx.stack([chosen_reward, rejected_reward])

        return loss, reward, num_tokens


from sillm.models import llama
from sillm.models import gemma
from sillm.models import qwen2
from sillm.models import mixtral
from sillm.models import phi
from sillm.models import starcoder2
from sillm.models import dbrx
from sillm.models import cohere


import re

import mlx.core as mx

def map_key(k):
    """
    Map weights key to MLX naming scheme.
    Args:
        k: Key to map.
    """
    if k.startswith("layers."):
        return k
    elif k.startswith("output."):
        return k
    elif k in ["tok_embeddings.weight", "norm.weight", "rope.freqs"]:
        return k
    elif k.startswith("model.embed_tokens."):
        return re.sub(r"^model\.embed_tokens\.", "tok_embeddings.", k)
    elif k.startswith("model.norm."):
        return re.sub(r"^model\.norm\.", "norm.", k)
    elif k.startswith("model.final_layernorm."):
        return re.sub(r"^model\.final_layernorm\.", "norm.", k)
    elif k.startswith("lm_head."):
        return re.sub(r"^lm_head\.", "output.", k)
    elif k.startswith("model.layers."):
        layer = k.split(".")[2]
        
        k = re.sub(r"^model\.layers\.", "layers.", k)

        k = re.sub(r"\.input_layernorm\.", ".attention_norm.", k)
        k = re.sub(r"\.post_attention_layernorm\.", ".ffn_norm.", k)

        k = re.sub(r"\.self_attn\.(q|k|v|o)_proj\.", r".attention.w\1.", k)
        k = re.sub(r"\.self_attn\.(q|k)_norm\.", r".attention.\1_norm.", k)
        k = re.sub(r"\.mlp\.gate_proj\.", ".feed_forward.w1.", k)
        k = re.sub(r"\.mlp\.down_proj\.", ".feed_forward.w2.", k)
        k = re.sub(r"\.mlp\.up_proj\.", ".feed_forward.w3.", k)

        # MoE
        k = re.sub(r"\.block_sparse_moe\.gate\.", ".feed_forward.gate.", k)
        k = re.sub(r"\.block_sparse_moe\.experts\.(\d+)\.w1.", r".feed_forward.experts.\1.w1.", k)
        k = re.sub(r"\.block_sparse_moe\.experts\.(\d+)\.w2.", r".feed_forward.experts.\1.w2.", k)
        k = re.sub(r"\.block_sparse_moe\.experts\.(\d+)\.w3.", r".feed_forward.experts.\1.w3.", k)

        # Phi mapping
        k = re.sub(r"\.self_attn\.dense\.", ".attention.wo.", k)
        k = re.sub(r"\.mlp\.fc1\.", ".feed_forward.w1.", k)
        k = re.sub(r"\.mlp\.fc2\.", ".feed_forward.w2.", k)

        # Starcoder2 mapping
        k = re.sub(r"\.mlp\.c_fc\.", ".feed_forward.w1.", k)
        k = re.sub(r"\.mlp\.c_proj\.", ".feed_forward.w2.", k)

        return k
    # GGUF keys
    elif k.startswith("output_norm."):
        return re.sub(r"^output_norm\.", "norm.", k)
    elif k.startswith("token_embd."):
        return re.sub(r"^token_embd\.", "tok_embeddings.", k)
    elif k.startswith("blk."):
        layer = k.split(".")[1]

        k = re.sub(r"^blk\.", "layers.", k)

        if k.endswith(".attn_norm.weight"):
            return f"layers.{layer}.attention_norm.weight"
        elif k.endswith(".ffn_norm.weight"):
            return f"layers.{layer}.ffn_norm.weight"

        k = re.sub(r"\.attn_(q|k|v)\.", r".attention.w\1.", k)
        k = re.sub(r"\.attn_output\.", r".attention.wo.", k)

        # MoE
        k = re.sub(r"\.ffn_gate_inp\.", ".feed_forward.gate.", k)
        k = re.sub(r"\.ffn_gate\.(\d+)\.", r".feed_forward.experts.\1.w1.", k)
        k = re.sub(r"\.ffn_down\.(\d+)\.", r".feed_forward.experts.\1.w2.", k)
        k = re.sub(r"\.ffn_up\.(\d+)\.", r".feed_forward.experts.\1.w3.", k)

        k = re.sub(r"\.ffn_gate\.", ".feed_forward.w1.", k)
        k = re.sub(r"\.ffn_down\.", ".feed_forward.w2.", k)
        k = re.sub(r"\.ffn_up\.", ".feed_forward.w3.", k)

        return k
    # DBRX keys
    elif k.startswith("transformer.wte."):
        return re.sub(r"^transformer\.wte\.", "tok_embeddings.", k)
    elif k.startswith("transformer.norm_f."):
        return re.sub(r"^transformer\.norm_f\.", "norm.", k)
    elif k.startswith("transformer.blocks."):
        layer = k.split(".")[2]

        k = re.sub(r"^transformer\.blocks\.", "layers.", k)
        k = re.sub(r"\.ffn\.router\.layer\.", ".feed_forward.gate.", k)

        k = re.sub(r"\.ffn\.experts\.(\d+)\.w1.", r".feed_forward.experts.\1.w1.", k)
        k = re.sub(r"\.ffn\.experts\.(\d+)\.w2.", r".feed_forward.experts.\1.w2.", k)
        k = re.sub(r"\.ffn\.experts\.(\d+)\.v1.", r".feed_forward.experts.\1.w3.", k)
        
        k = re.sub(r"\.norm_attn_norm\.attn\.Wqkv\.", ".attention.wqkv.", k)
        k = re.sub(r"\.norm_attn_norm\.attn\.out_proj\.", ".attention.wo.", k)
        k = re.sub(r"\.norm_attn_norm\.norm_1\.", ".attention_norm.", k)
        k = re.sub(r"\.norm_attn_norm\.norm_2\.", ".ffn_norm.", k)

        return k
    
    return None

def map_keys(keys):
    """
    Map weights keys to MLX naming scheme.
    Args:
        keys: Keys to map.
    """
    result = {}
    for k in keys:
        result[k] = map_key(k)

    return result

def map_config(config):
    """
    Map config to MLX naming scheme.
    Args:
        config: Configuration to map.
    """
    result = {}

    mlx_keys = [
        "model_type",
        "dim",
        "n_layers",
        "n_heads",
        "head_dim",
        "hidden_dim",
        "n_kv_heads",
        "norm_eps",
        "vocab_size",
        "max_position_embeddings",
        "rope_theta",
        "rope_scaling",
        "bos_token_id",
        "eos_token_id",
        "pad_token_id",
        "quantization",
        "moe",
        "tie_word_embeddings",
        "clip_qkv",
        "use_qk_norm",
        "logit_scale",
    ]
    for key in mlx_keys:
        if key in config:
            result[key] = config[key]

    key_map = {
        "hidden_size": "dim",
        "num_hidden_layers": "n_layers",
        "num_attention_heads": "n_heads",
        "intermediate_size": "hidden_dim",
        "num_key_value_heads": "n_kv_heads",
        "n_heads": "n_kv_heads",
        "rms_norm_eps": "norm_eps",
        "layer_norm_eps": "norm_eps",
        "norm_epsilon": "norm_eps",
        "layer_norm_bias": "norm_bias",
        # GGUF metadata: https://github.com/ggerganov/ggml/blob/master/docs/gguf.md
        "llama.embedding_length": "dim",
        "llama.block_count": "n_layers",
        "llama.attention.head_count": "n_heads",
        "llama.feed_forward_length": "hidden_dim",
        "llama.attention.head_count_kv": "n_kv_heads",
        "llama.attention.layer_norm_rms_epsilon": "norm_eps",
        "llama.rope.freq_base": "rope_theta",
        "llama.context_length": "max_position_embeddings",
        "tokenizer.ggml.bos_token_id": "bos_token_id",
        "tokenizer.ggml.eos_token_id": "eos_token_id",
        # DBRX
        "d_model": "dim",
    }

    for key in key_map:
        if key in config and key_map[key] not in result:
            value = config[key]
            if isinstance(value, mx.array):
                value = value.item()
            result[key_map[key]] = value

    if "general.architecture" in config:
        result["model_type"] = config["general.architecture"]
        if "general.name" in config and "mixtral" in config["general.name"].lower():
            result["model_type"] = "mixtral"
    if "tokenizer.ggml.tokens" in config:
        result["vocab_size"] = len(config["tokenizer.ggml.tokens"])
    if result["vocab_size"] <= 0:
        del(result["vocab_size"])
    if "llama.expert_count" in config and "llama.expert_used_count" in config:
        result["moe"] = {
            "num_experts": config["llama.expert_count"].item(),
            "num_experts_per_tok": config["llama.expert_used_count"].item()
        }

    # DBRX config
    if "attn_config" in config:
        attn_config = config["attn_config"]

        if "kv_n_heads" in attn_config:
            result["n_kv_heads"] = attn_config["kv_n_heads"]
        if "clip_qkv" in attn_config:
            result["clip_qkv"] = attn_config["clip_qkv"]
        if "rope_theta" in attn_config:
            result["rope_theta"] = attn_config["rope_theta"]
    if "ffn_config" in config:
        ffn_config = config["ffn_config"]

        if "ffn_hidden_size" in ffn_config:
            result["hidden_dim"] = ffn_config["ffn_hidden_size"]
        if "moe_num_experts" in ffn_config and "moe_top_k" in ffn_config:
            result["moe"] = {
                "num_experts": ffn_config["moe_num_experts"],
                "num_experts_per_tok": ffn_config["moe_top_k"]
            }

    # Calculate head_dim if not provided
    if "head_dim" not in result:
        if "dim" in result and "n_heads" in result:
            result["head_dim"] = result["dim"] // result["n_heads"]

    return result


from sillm.core.loader import load
from sillm.training.dataset import load_dataset
from sillm.core.template import init_template

from sillm.core.llm import LLM
from sillm.core.conversation import Conversation, AutoConversation
from sillm.core.template import Template, AutoTemplate
from sillm.training.lora import TrainableLoRA
from sillm.training.dpo import TrainableDPO
from sillm.training.dataset import DatasetCompletion, DatasetInstruct, DatasetDPO
from sillm.models.args import ModelArgs

__version__ = "0.1.0"


import os
import argparse

import mlx.core as mx

import sillm
import sillm.utils as utils

if __name__ == "__main__":
    # Parse commandline arguments
    parser = argparse.ArgumentParser(description="Interface for training SiLLM models with LoRA/QLoRA.")
    parser.add_argument("model", type=str, help="The directory or file for the base model (MLX, Torch, GGUF)")
    parser.add_argument("-d", "--chdir", default=None, type=str, help="Change working directory")
    parser.add_argument("-c", "--config", default=None, type=str, help="Load YAML configuration file for training")
    parser.add_argument("-t", "--train", default=None, type=str, help="Train the model with training dataset in the file/directory")
    parser.add_argument("-a", "--input_adapters", default=None, type=str, help="Load adapter weights from .safetensors file")
    parser.add_argument("-o", "--output_dir", default=None, type=str, help="Output directory for adapter weights")
    parser.add_argument("-s", "--save_checkpoints", default=False, action="store_true", help="Save model checkpoints to output directory")
    parser.add_argument("-m", "--save_merge", default=None, type=str, help="Save merged model weights to .safetensors file")
    parser.add_argument("-g", "--grad_checkpoint", default=False, action="store_true", help="Use gradient checkpointing")
    parser.add_argument("--max_length", default=1024, type=int, help="Max token length per training dataset entry (default: 1024)")
    parser.add_argument("--template", type=str, default=None, help="Chat template (chatml, llama-2, alpaca, etc.)")
    parser.add_argument("--layers", default=-1, type=int, help="Layers to use for LoRA (default: -1 for all layers)")
    parser.add_argument("--target_modules", default="query_value", type=str, help="Target modules to use for LoRA: query_value, all_linear")
    parser.add_argument("--rank", default=8, type=int, help="Rank to use for LoRA (default: 8)")
    parser.add_argument("--learning_rate", default=1e-5, type=float, help="Learning rate (default: 1e-5)")
    parser.add_argument("--epochs", default=1, type=int, help="Number of epochs (default: 1)")
    parser.add_argument("--iterations", default=0, type=int, help="Number of iterations per epoch (default: dataset size)")
    parser.add_argument("--batch_size", default=4, type=int, help="Size of training batches (default: 4)")
    parser.add_argument("--report_steps", default=10, type=int, help="Number of batch iterations per training report (default: 10)")
    parser.add_argument("--eval_steps", default=100, type=int, help="Number of batch iterations per evaluation (default: 100)")
    parser.add_argument("--validation_samples", default=40, type=int, help="Number of validation_samples (default: 40)")
    parser.add_argument("--seed", default=0, type=int, help="Seed for randomization (default: 0)")
    parser.add_argument("--plot", default=None, type=str, help="Create a loss plot and save it to the specified file")
    parser.add_argument("-q4", default=False, action="store_true", help="Quantize the model to 4 bits")
    parser.add_argument("-q8", default=False, action="store_true", help="Quantize the model to 8 bits")
    parser.add_argument("-v", "--verbose", default=1, action="count", help="Increase output verbosity")
    args = parser.parse_args()

    # Change working directory
    if args.chdir is not None:
        os.chdir(args.chdir)

    # Load YAML configuration file
    if args.config is not None:
        utils.load_yaml(args.config, args)
    
    # Initialize logging
    log_level = 40 - (10 * args.verbose) if args.verbose > 0 else 0
    logger = utils.init_logger(log_level)

    # Log commandline arguments
    if log_level <= 10:
        utils.log_arguments(args.__dict__)

    # Load base model
    model = sillm.load(args.model)
    
    # Set random seed
    if args.seed >= 0:
        utils.seed(args.seed)

    # Quantize model
    if args.q4 is True:
        model.quantize(bits=4)
    elif args.q8 is True:
        model.quantize(bits=8)

    # Initialize trainable model
    model = sillm.TrainableLoRA.from_model(model)
    
    # Initialize LoRA layers
    lora_config = {
        "num_layers":       args.layers,
        "target_modules":   args.target_modules,
        "rank":             args.rank
    }
    model.init_lora(**lora_config)

    if args.input_adapters is not None:
        # Load adapter file
        model.load_adapters(args.input_adapters)

    # Initialize plot
    if args.plot is not None:
        plot = utils.Plot()

    # Set conversation template
    if args.template:
        template = sillm.Template(model.tokenizer, template=args.template)
    else:
        template = None

    # Log memory usage
    utils.log_memory_usage()

    if args.train is not None:
        # Load training dataset
        dataset_config = {
            "template": template,
            "max_length": args.max_length
        }
        dataset_training, dataset_validation, dataset_test = sillm.load_dataset(model.tokenizer, args.train, **dataset_config)

        def report_callback(i, loss):
            if args.plot is not None:
                plot.add_train_loss(i, loss)
        
        def eval_callback(i, val_loss):
            if args.plot is not None:
                plot.add_valid_loss(i, val_loss)
                plot.save(args.plot)

            if i > 1 and args.save_checkpoints and args.output_dir is not None:
                fpath_ckpt = model.save_checkpoint(args.output_dir, i)

                return f"Saved checkpoint to {fpath_ckpt}"
            
        if args.output_dir is not None:
            model.save_lora_config(args.output_dir)

        # Model training
        training_config = {
            "grad_checkpoint":      args.grad_checkpoint,
            "batch_size":           args.batch_size,
            "learning_rate":        args.learning_rate,
            "epochs":               args.epochs,
            "iterations":           args.iterations,
            "report_steps":         args.report_steps,
            "eval_steps":           args.eval_steps,
            "validation_samples":   args.validation_samples,
        }
        model.train(dataset_training,
                    dataset_validation,
                    report_callback=report_callback,
                    eval_callback=eval_callback,
                    **training_config)

    if args.plot is not None:
        plot.save(args.plot)

    if args.save_checkpoints:
        # Save final checkpoint
        model.save_checkpoint(args.output_dir)

    if args.save_merge:
        # Merge LoRA layers back into model
        model.merge_and_unload_lora()

        # Save merged weights
        model.save_weights(args.save_merge)


from typing import Optional, Tuple

import mlx.core as mx
import mlx.nn as nn

from sillm.models.base import BaseModel
from sillm.models.args import ModelArgs
import sillm.models.llama as llama

class LayerNorm2D(nn.Module):
    def __init__(self, d1, d2, eps):
        super().__init__()

        self.weight = mx.zeros((d1, d2))
        self.eps = eps

    def __call__(self, x):
        return self.weight * mx.fast.layer_norm(x, None, None, self.eps)
    
class Attention(nn.Module):
    """
    Multi-head attention module.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        super().__init__()
        self.args = args

        self.n_heads = args.n_heads
        self.n_kv_heads = args.n_kv_heads
        self.use_qk_norm = args.use_qk_norm

        self.scale = self.args.head_dim ** -0.5

        self.wq = nn.Linear(args.dim, args.n_heads * args.head_dim, bias=False)
        self.wk = nn.Linear(args.dim, args.n_kv_heads * args.head_dim, bias=False)
        self.wv = nn.Linear(args.dim, args.n_kv_heads * args.head_dim, bias=False)
        self.wo = nn.Linear(args.n_heads * args.head_dim, args.dim, bias=False)
        
        if self.use_qk_norm:
            self.q_norm = LayerNorm2D(self.n_heads, args.head_dim, eps=args.norm_eps)
            self.k_norm = LayerNorm2D(self.n_kv_heads, args.head_dim, eps=args.norm_eps)

        self.rope = nn.RoPE(args.head_dim, traditional=True, base=args.rope_theta)

    def __call__(self,
                 x: mx.array,
                 mask: Optional[mx.array] = None,
                 cache: Optional[Tuple[mx.array, mx.array]] = None,
                 ) -> mx.array:
        B, L, _ = x.shape

        queries, keys, values = self.wq(x), self.wk(x), self.wv(x)

        queries = queries.reshape(B, L, self.n_heads, -1)
        keys = keys.reshape(B, L, self.n_kv_heads, -1)
        if self.use_qk_norm:
            queries = self.q_norm(queries)
            keys = self.k_norm(keys)
        queries = queries.transpose(0, 2, 1, 3)
        keys = keys.transpose(0, 2, 1, 3)
        values = values.reshape(B, L, self.n_kv_heads, -1).transpose(0, 2, 1, 3)

        if cache is not None:
            key_cache, value_cache = cache
            queries = self.rope(queries, offset=key_cache.shape[2])
            keys = self.rope(keys, offset=key_cache.shape[2])
            keys = mx.concatenate([key_cache, keys], axis=2)
            values = mx.concatenate([value_cache, values], axis=2)
        else:
            queries = self.rope(queries)
            keys = self.rope(keys)

        output = mx.fast.scaled_dot_product_attention(
            queries, keys, values, scale=self.scale, mask=mask
        )
        output = output.transpose(0, 2, 1, 3).reshape(B, L, -1)

        return self.wo(output), (keys, values)

class TransformerBlock(nn.Module):
    """
    Transformer block.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        super().__init__()
        self.args = args
        
        self.n_heads = args.n_heads
        self.dim = args.dim
        
        self.attention = Attention(args=args)
        self.feed_forward = llama.FeedForward(args=args)
        self.attention_norm = nn.LayerNorm(args.dim, eps=args.norm_eps, bias=args.norm_bias)

    def forward(
            self,
            x: mx.array,
            mask: Optional[mx.array] = None,
            cache: Optional[Tuple[mx.array, mx.array]] = None,
            ) -> mx.array:
        """
        Args:
            x: Input tensor.
            mask: Mask tensor.
            cache: Cache from previous forward pass.
        Returns:
            Output tensor and cache.
        """
        h = self.attention_norm(x)
        a, cache = self.attention(h, mask, cache)
        r = self.feed_forward(h)
        out = a + r + x
        
        return out, cache

class Model(BaseModel):
    """
    Cohere model wrapper.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        # Override RoPE settings
        args.rope_traditional = True

        super().__init__(args)
        self.args = args

        self.n_layers = args.n_layers
        self.vocab_size = args.vocab_size
        self.logit_scale = args.logit_scale
        
        self.tok_embeddings = nn.Embedding(args.vocab_size, args.dim)
        self.layers = [TransformerBlock(args=args) for _ in range(args.n_layers)]
        self.norm = nn.LayerNorm(args.dim, eps=args.norm_eps, bias=args.norm_bias)

    def __call__(self,
                 inputs: mx.array,
                 cache = None
                 ):
        """
        Args:
            inputs: Input tokens.
            cache: Cache from previous forward pass.
        Returns:
            Output logits and cache.
        """
        h = self.tok_embeddings(inputs)

        mask = None
        if h.shape[1] > 1:
            mask = nn.MultiHeadAttention.create_additive_causal_mask(h.shape[1])
            mask = mask.astype(h.dtype)

        if cache is None:
            cache = [None] * len(self.layers)

        for e, layer in enumerate(self.layers):
            h, cache[e] = layer.forward(h, mask, cache[e])

        out = self.tok_embeddings.as_linear(self.norm(h))
        
        return out * self.logit_scale, cache


import argparse

import sillm
import sillm.utils as utils

if __name__ == "__main__":
    # Parse commandline arguments
    parser = argparse.ArgumentParser(description="A simple CLI for generating text with SiLLM.")
    parser.add_argument("input", type=str, help="The input model directory or file")
    parser.add_argument("output", type=str, help="The output model directory or file")
    parser.add_argument("-a", "--input_adapters", default=None, type=str, help="Load LoRA adapter weights from .safetensors file")
    parser.add_argument("-r", "--remap", default=False, action="store_true", help="Remap weights keys to native SiLLM format")
    parser.add_argument("-q4", default=False, action="store_true", help="Quantize the model to 4 bits")
    parser.add_argument("-q8", default=False, action="store_true", help="Quantize the model to 8 bits")
    parser.add_argument("-v", "--verbose", default=1, action="count", help="Increase output verbosity")
    args = parser.parse_args()

    # Initialize logging
    log_level = 40 - (10 * args.verbose) if args.verbose > 0 else 0
    logger = utils.init_logger(log_level)

    # Log commandline arguments
    if log_level <= 10:
        utils.log_arguments(args.__dict__)

    # Load model
    model = sillm.load(args.input)

    if args.input_adapters is not None:
        # Convert model to trainable
        model = sillm.TrainableLoRA.from_model(model)

        lora_config = model.load_lora_config(args.input_adapters)

        # Initialize LoRA layers
        model.init_lora(**lora_config)

        # Load and merge adapter file
        model.load_adapters(args.input_adapters)
        model.merge_and_unload_lora()

    # Quantize model
    if args.q4 is True:
        model.quantize(bits=4)
    elif args.q8 is True:
        model.quantize(bits=8)

    # Disable mapping to old keys
    if args.remap:
        model._mapping = None

    # Save model
    model.save(args.output)


import mlx.core as mx
import mlx.nn as nn

from sillm.models.base import BaseModel
from sillm.models.args import ModelArgs
import sillm.models.llama as llama

class Attention(llama.Attention):
    """
    Multi-head attention module.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        nn.Module.__init__(self)
        self.args = args

        self.n_heads: int = args.n_heads
        self.n_kv_heads: int = args.n_kv_heads

        self.scale = self.args.head_dim ** -0.5

        self.wq = nn.Linear(args.dim, args.n_heads * args.head_dim, bias=True)
        self.wk = nn.Linear(args.dim, args.n_kv_heads * args.head_dim, bias=True)
        self.wv = nn.Linear(args.dim, args.n_kv_heads * args.head_dim, bias=True)
        self.wo = nn.Linear(args.n_heads * args.head_dim, args.dim, bias=False)

        self.rope = nn.RoPE(args.head_dim,
                            traditional=args.rope_traditional,
                            base=args.rope_theta
                            )
        
class TransformerBlock(llama.TransformerBlock):
    """
    Transformer block.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        nn.Module.__init__(self)
        self.args = args

        self.n_heads = args.n_heads
        self.dim = args.dim
        
        self.attention = Attention(args=args)
        self.feed_forward = llama.FeedForward(args=args)
        self.attention_norm = nn.RMSNorm(args.dim, eps=args.norm_eps)
        self.ffn_norm = nn.RMSNorm(args.dim, eps=args.norm_eps)

########
# References:
# https://github.com/huggingface/transformers/blob/main/src/transformers/models/gemma/modeling_gemma.py
########
class Model(llama.Model):
    """
    Qwen 2 model wrapper.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        BaseModel.__init__(self, args)
        self.args = args

        self.n_layers = args.n_layers
        self.vocab_size = args.vocab_size
        
        self.tok_embeddings = nn.Embedding(args.vocab_size, args.dim)
        self.layers = [TransformerBlock(args=args) for _ in range(args.n_layers)]
        self.norm = nn.RMSNorm(args.dim, eps=args.norm_eps)

        if args.tie_word_embeddings:
            self.output = None
        else:
            self.output = nn.Linear(args.dim, args.vocab_size, bias=False)

    def __call__(self,
                 inputs: mx.array,
                 cache = None
                 ):
        """
        Args:
            inputs: Input tokens.
            cache: Cache from previous forward pass.
        Returns:
            Output logits and cache.
        """
        h = self.tok_embeddings(inputs)

        mask = None
        if h.shape[1] > 1:
            mask = nn.MultiHeadAttention.create_additive_causal_mask(h.shape[1])
            mask = mask.astype(h.dtype)

        if cache is None:
            cache = [None] * len(self.layers)

        for e, layer in enumerate(self.layers):
            h, cache[e] = layer.forward(h, mask, cache[e])

        if self.output is None:
            out = self.tok_embeddings.as_linear(self.norm(h))
        else:
            out = self.output(self.norm(h))

        return out, cache


import logging
import pathlib
import json

import mlx.core as mx

from .mapping import map_key

logger = logging.getLogger("sillm")

def quantize_files(input_path: str,
                   output_path: str,
                   group_size: int = 32,
                   bits: int = 4
                   ):
    """
    Quantize weights files.
    Args:
        input_path: Path to load weights.
        output_path: Path to save quantized weights.
    """
    input_path = pathlib.Path(input_path)
    output_path = pathlib.Path(output_path)
    output_path.mkdir(parents=True, exist_ok=True)

    weights_files = sorted(list(input_path.glob("*.safetensors")))
    if len(weights_files) == 0:
        raise ValueError("No weights files found")
    
    def save_shard(shard, shard_path):
        mx.save_safetensors(shard_path, shard)
        logger.debug(f"Saved quantized shard to {shard_path}")
    
    shard = {}
    weight_map = {}
    original_size = 0
    total_size = 0
    for weights_path in weights_files:
        logger.debug(f"Loading model weights file {weights_path}")

        weights = mx.load(str(weights_path))
        for key, weight in weights.items():
            original_size += weight.nbytes
            mlx_key = map_key(key)

            # Checking if weight should be quantized
            quantize_weight = key.endswith("weight") and len(weight.shape) > 1
            if weight.shape[0] == 8 or ".gate." in mlx_key:
                logger.debug(f"Skipping quantization for MoE gate: {key}")
                quantize_weight = False
            elif ".k_norm." in mlx_key or ".q_norm." in mlx_key:
                logger.debug(f"Skipping quantization for attention norm: {key}")
                quantize_weight = False
            elif len(weight.shape) > 1 and weight.shape[1] < 512:
                if bits == 4 and weight.shape[1] < 256:
                    logger.debug(f"Skipping quantization for small weight: {key}")
                    quantize_weight = False
                if bits == 8 and weight.shape[1] < 128:
                    logger.debug(f"Skipping quantization for small weight: {key}")
                    quantize_weight = False

            if quantize_weight:
                weight, scales, biases = mx.quantize(weight, group_size, bits)
                quant_size = weight.nbytes + scales.nbytes + biases.nbytes
                total_size += quant_size

                key_scales = key.replace(".weight", ".scales")
                key_biases = key.replace(".weight", ".biases")

                shard[key_scales] = scales
                shard[key_biases] = biases
            shard[key] = weight
            weight_map[key] = weights_path.name

        shard_path = str(output_path / weights_path.name)
        save_shard(shard, shard_path)
        shard = {}

    logger.debug(f"Quantization reduced weights size: {original_size//1024//1024:,} MB => {total_size//1024//1024:,} MB")

    index_path = str(output_path / "model.safetensors.index.json")
    with open(index_path, "w") as f:
        index_data = {
            "metadata": {
                "total_size": total_size,
            },
            "weight_map": weight_map
        }

        f.write(json.dumps(index_data, indent=4))
        logger.debug(f"Saved weight index to {index_path}")


import logging

from sillm.core.template import Template, AutoTemplate

logger = logging.getLogger("sillm")

class Conversation(object):
    """
    Wrapper for conversation templates.
    """
    def __init__(self,
                template: Template,
                system_prompt: str = None
                ):
        self.template = template
        self.system_prompt = system_prompt

        self.messages = []

    def apply_chat_template(self,
                            add_generation_prompt: bool = False
                            ):
        return self.template.apply_chat_template(messages=self.messages, add_generation_prompt=add_generation_prompt)

    def __str__(self):
        """
        Return formatted string using the template.
        Returns:
            Formatted conversation string.
        """
        return self.apply_chat_template()
        
    def clear(self):
        """
        Clear conversation.
        """
        self.messages = []

    def load_messages(self,
                      messages: list
                      ):
        """
        Load messages into conversation.
        Args:
            messages: List of messages.
        """
        self.messages = messages

    def add_message(self,
                    content: str,
                    role: str,
                    add_generation_prompt: bool = True
                    ):
        if len(self.messages) == 0 and self.system_prompt is not None:
            # Add system message
            msg_system = {
                "role": "system",
                "content": self.system_prompt
            }
            self.messages.append(msg_system)
        msg = {
            "role": role,
            "content": content
        }
        self.messages.append(msg)

        return self.apply_chat_template(add_generation_prompt=add_generation_prompt)

    def add_user(self,
                 content: str,
                 add_generation_prompt: bool = True
                 ):
        """
        Add user message to the conversation.
        Args:
            content: User prompt.
            add_generation_prompt: Whether to add generation prompt.
        Returns:
            Formatted conversation string.
        """
        return self.add_message(content=content, role="user", add_generation_prompt=add_generation_prompt)
    
    def add_assistant(self,
                     content: str
                     ):
        """
        Add assistant message to the conversation.
        Args:
            content: Assistant response.
        Returns:
            Formatted conversation string.
        """
        return self.add_message(content=content, role="assistant")
    
class AutoConversation(Conversation):
    """
    Wrapper for tokenizers with built-in chat templates.
    """
    def __init__(self,
                tokenizer,
                system_prompt: str = None):
        self.tokenizer = tokenizer
        self.system_prompt = system_prompt
        self.messages = []

        logger.info(f"Initialized built-in conversation template from tokenizer")

    def apply_chat_template(self,
                            add_generation_prompt: bool = False
                            ):
        return self.tokenizer.apply_chat_template(self.messages, tokenize=False, add_generation_prompt=add_generation_prompt)


import sys
import os
import logging

import mlx.core as mx

logger = logging.getLogger("sillm")

def init_logger(log_level,
                fmt = "%(asctime)s %(levelname)s %(message)s",
                datefmt = "%Y-%m-%d %H:%M:%S",
                add_stdout = True
                ):
    """
    Initialize the logger.
    Args:
        log_level: The log level.
        fmt: The log format.
        datefmt: The date format.
    Returns:
        The logger.
    """
    # Set log level
    logger.setLevel(log_level)

    if add_stdout:
        # Initialize formatter and handler
        formatter = logging.Formatter(fmt, datefmt=datefmt)
        handler = logging.StreamHandler(sys.stdout)
        handler.setFormatter(formatter)
        logger.addHandler(handler)

    return logger

def log_arguments(args):
    """
    Log arguments.
    
    Args:
        args: The parsed arguments.
    """
    logger = logging.getLogger("sillm")

    for key, value in args.items():
        logger.debug(f"Argument: {key} = {value}")

def log_memory_usage():
    peak_memory = mx.metal.get_peak_memory()
    system_memory = os.sysconf("SC_PAGE_SIZE") * os.sysconf("SC_PHYS_PAGES")
    memory_usage = peak_memory / system_memory

    logger.debug(f"Peak memory: {(peak_memory // (1024 ** 2)):,} MB ({memory_usage:.2%} of system memory)")


import pathlib
import json
import logging

import numpy as np

import mlx.core as mx

logger = logging.getLogger("sillm")

class Dataset:
    """
    Dataset wrapper.
    """
    def __init__(self,
                 entries,
                 tokenizer,
                 template = None,
                 max_length: int = 4096
                 ):
        """
        Args:
            tokenizer: Tokenizer to use.
            dataset_path: Path to dataset file.
            key: Key to use for text.
            max_length: Max token length per training entry.
        """
        raise NotImplementedError("Class Dataset is used for inheritance only")
                 
    def __getitem__(self, idx: int):
        return self._data[idx]

    def __len__(self):
        return len(self._data)
    
class DatasetCompletion(Dataset):
    """
    Completion dataset wrapper.
    """
    def __init__(self,
                 entries,
                 tokenizer,
                 template = None,
                 max_length: int = 4096
                 ):
        """
        Args:
            tokenizer: Tokenizer to use.
            dataset_path: Path to dataset file.
            key: Key to use for text.
            max_length: Max token length per training entry.
        """
        self._type_id = "Completion"
        self._data = []
        if tokenizer.pad_id is not None and tokenizer.pad_id >= 0:
            self.pad_id = tokenizer.pad_id
        else:
            self.pad_id = tokenizer.eos_id

        key = "text"

        num_outsized = 0
        for entry in entries:
            tokens = tokenizer.encode(entry[key], eos=True)

            if len(tokens) < max_length:
                self._data.append(tokens)
            else:
                num_outsized += 1

        if num_outsized > 0:
            logger.debug(f"Removed {num_outsized} entries from dataset due to max. length {max_length}")
    
    ########
    # Based on mlx-examples:
    # https://github.com/ml-explore/mlx-examples/blob/e74889d0fa0fb49d95bfdf6a1dcad907713eb50e/lora/lora.py#L166
    ########
    def iterate_batches(self,
                        batch_size: int,
                        train: bool = False):
        """
        Iterate over batches.
        Args:
            batch_size: Batch size.
            train: Whether to train.
        """
        # Shuffle indices
        while True:
            indices = np.arange(len(self._data))
            if train:
                indices = np.random.permutation(indices)

            # Collect batches from dataset
            for i in range(0, len(indices) - batch_size + 1, batch_size):
                batch = [self._data[indices[i+j]] for j in range(batch_size)]
                lengths = [len(x) for x in batch]

                # Pad to the max length
                batch_arr = np.full((batch_size, max(lengths)), self.pad_id, np.int32)
                for j in range(batch_size):
                    batch_arr[j, : lengths[j]] = batch[j]
                batch = mx.array(batch_arr)

                inputs = batch[:, :-1]
                targets = batch[:, 1:]

                lengths = mx.array(lengths)
                loss_masks = mx.arange(inputs.shape[1])[None, :] < lengths[:, None]

                yield inputs, targets, loss_masks

            if not train:
                break
    
class DatasetMessages(Dataset):
    """
    Messages dataset wrapper.
    """
    def __init__(self,
                 entries,
                 tokenizer,
                 template,
                 max_length: int = 4096
                 ):
        """
        Args:
            tokenizer: Tokenizer to use.
            dataset_path: Path to dataset file.
            key: Key to use for text.
            max_length: Max token length per training entry.
        """
        if template is None:
            raise ValueError("Template must be specified for Messages dataset")

        self._type_id = "Messages"
        self._data = []
        if tokenizer.pad_id is not None and tokenizer.pad_id >= 0:
            self.pad_id = tokenizer.pad_id
        else:
            self.pad_id = tokenizer.eos_id

        messages_key = "messages"

        num_outsized = 0
        for entry in entries:
            messages = entry[messages_key]

            text_prompt = template.apply_chat_template(messages[:-1])
            tokens_prompt = tokenizer.encode(text_prompt)
            
            text = template.apply_chat_template(messages)
            tokens = tokenizer.encode(text, eos=True)

            if len(tokens) < max_length:
                self._data.append((tokens_prompt, tokens))
            else:
                num_outsized += 1

        if num_outsized > 0:
            logger.debug(f"Removed {num_outsized} entries from dataset due to max. length {max_length}")

    def iterate_batches(self,
                        batch_size: int,
                        train: bool = False):
        """
        Iterate over batches.
        Args:
            batch_size: Batch size.
            train: Whether to train.
        """
        # Shuffle indices
        while True:
            indices = np.arange(len(self._data))
            if train:
                indices = np.random.permutation(indices)

            # Collect batches from dataset
            for i in range(0, len(indices) - batch_size + 1, batch_size):
                prompts = [self._data[indices[i+j]][0] for j in range(batch_size)]
                batch = [self._data[indices[i+j]][1] for j in range(batch_size)]

                # Pad to the max length
                lengths = [len(x) for x in batch]
                batch_arr = np.full((batch_size, max(lengths)), self.pad_id, np.int32)
                for j in range(batch_size):
                    batch_arr[j, : lengths[j]] = batch[j]
                batch = mx.array(batch_arr)

                inputs = batch[:, :-1]
                targets = batch[:, 1:]

                lengths = mx.array(lengths)
                prompt_lengths = mx.array([len(x) for x in prompts])
                loss_masks = mx.logical_and(
                    mx.arange(inputs.shape[1])[None, :] < lengths[:, None],
                    mx.arange(inputs.shape[1])[None, :] > prompt_lengths[:, None]
                )

                yield inputs, targets, loss_masks

            if not train:
                break

class DatasetInstruct(DatasetMessages):
    """
    Q&A dataset wrapper.
    """
    def __init__(self,
                 entries,
                 tokenizer,
                 template = None,
                 max_length: int = 4096
                 ):
        """
        Args:
            tokenizer: Tokenizer to use.
            dataset_path: Path to dataset file.
            key: Key to use for text.
            max_length: Max token length per training entry.
        """
        self._type_id = "Instruct"
        self._data = []
        if tokenizer.pad_id is not None and tokenizer.pad_id >= 0:
            self.pad_id = tokenizer.pad_id
        else:
            self.pad_id = tokenizer.eos_id

        prompt_key = "prompt"
        response_key = "response"

        num_outsized = 0
        for entry in entries:
            prompt = entry[prompt_key]
            response = entry[response_key]
            if template is None:
                tokens_prompt = tokenizer.encode(prompt)
                tokens = tokenizer.encode(prompt + response, eos=True)
            else:
                messages_prompt = [
                    { "role": "user", "content": prompt}
                ]
                tokens_prompt = tokenizer.encode(template.apply_chat_template(messages_prompt))

                messages_response = [
                    { "role": "user", "content": prompt},
                    { "role": "assistant", "content": response }
                ]
                tokens = tokenizer.encode(template.apply_chat_template(messages_response), eos=True)

            if len(tokens) < max_length:
                self._data.append((tokens_prompt, tokens))
            else:
                num_outsized += 1

        if num_outsized > 0:
            logger.debug(f"Removed {num_outsized} entries from dataset due to max. length {max_length}")

class DatasetDPO(Dataset):
    """
    DPO dataset wrapper.
    """
    def __init__(self,
                 entries,
                 tokenizer,
                 template = None,
                 max_length: int = 4096
                 ):
        """
        Args:
            tokenizer: Tokenizer to use.
            dataset_path: Path to dataset file.
            key: Key to use for text.
            max_length: Max token length per training entry.
        """
        self._type_id = "DPO"
        self._data = []
        if tokenizer.pad_id is not None and tokenizer.pad_id >= 0:
            self.pad_id = tokenizer.pad_id
        else:
            self.pad_id = tokenizer.eos_id

        prompt_key = "prompt"
        chosen_key = "chosen"
        rejected_key = "rejected"

        num_outsized = 0
        for entry in entries:
            prompt = entry[prompt_key]
            chosen = entry[chosen_key]
            rejected = entry[rejected_key]
            if template is None:
                tokens_prompt = tokenizer.encode(prompt)
                tokens_chosen = tokenizer.encode(prompt+chosen, eos=True)
                tokens_rejected = tokenizer.encode(prompt+rejected, eos=True)
            else:
                messages_prompt = [
                    { "role": "user", "content": prompt}
                ]
                tokens_prompt = tokenizer.encode(template.apply_chat_template(messages_prompt))

                messages_chosen = [
                    { "role": "user", "content": prompt},
                    { "role": "assistant", "content": chosen}
                ]
                tokens_chosen = tokenizer.encode(template.apply_chat_template(messages_chosen), eos=True)

                messages_rejected = [
                    { "role": "user", "content": prompt},
                    { "role": "assistant", "content": rejected}
                ]
                tokens_rejected = tokenizer.encode(template.apply_chat_template(messages_rejected), eos=True)

            if len(tokens_chosen) < max_length and len(tokens_rejected) < max_length:
                self._data.append((tokens_prompt, tokens_chosen, tokens_rejected))
            else:
                num_outsized += 1
        
        if num_outsized > 0:
            logger.debug(f"Removed {num_outsized} entries from dataset due to max. length {max_length}")

    def iterate_batches(self,
                        batch_size: int,
                        train: bool = False):
        """
        Iterate over batches.
        Args:
            batch_size: Batch size (unused).
            train: Whether to train.
        """
        # Shuffle indices
        while True:
            indices = np.arange(len(self._data))
            if train:
                indices = np.random.permutation(indices)

            # Collect batches from dataset
            for i in range(0, len(indices) - batch_size + 1, batch_size):
                batch = [self._data[indices[i+j]] for j in range(batch_size)]
                chosen_lengths = [len(x[1]) for x in batch]
                rejected_lengths = [len(x[2]) for x in batch]

                # Pad to the max length
                chosen = np.full((batch_size, max(chosen_lengths)), self.pad_id, np.int32)
                rejected = np.full((batch_size, max(rejected_lengths)), self.pad_id, np.int32)
                for j in range(batch_size):
                    chosen[j, : chosen_lengths[j]] = batch[j][1]
                    rejected[j, : rejected_lengths[j]] = batch[j][2]

                # Mask prompt and padding tokens
                chosen_lengths = mx.array(chosen_lengths)
                rejected_lengths = mx.array(rejected_lengths)
                prompt_lengths = mx.array([len(x[0]) for x in batch])
                chosen_masks = mx.logical_and(
                    mx.arange(chosen.shape[1] - 1)[None, :] < chosen_lengths[:, None],
                    mx.arange(chosen.shape[1] - 1)[None, :] > prompt_lengths[:, None]
                )
                rejected_masks = mx.logical_and(
                    mx.arange(rejected.shape[1] - 1)[None, :] < rejected_lengths[:, None],
                    mx.arange(rejected.shape[1] - 1)[None, :] > prompt_lengths[:, None]
                )
                
                yield mx.array(chosen), mx.array(rejected), chosen_masks, rejected_masks

            if not train:
                break
    
def load_jsonl(fpath,
               shuffle: bool = True
               ):
    entries = []
    with open(fpath, "r") as f:
        for line in f:
            entry = json.loads(line)
            entries.append(entry)

    if shuffle:
        np.random.shuffle(entries)

    return entries
    
def load_dataset(tokenizer,
                 dataset_path,
                 train_split: float = 0.9,
                 valid_split: float = 0.05,
                 test_split: float = 0.05,
                 template = None,
                 max_length: int = 4096,
                 shuffle: bool = True
                 ):
    dataset_path = pathlib.Path(dataset_path)

    def guess_type(entry):
        if "text" in entry:
            return DatasetCompletion
        elif "messages" in entry:
            return DatasetMessages
        elif "prompt" in entry and "chosen" in entry and "rejected" in entry:
            return DatasetDPO
        elif "prompt" in entry and "response" in entry:
            return DatasetInstruct
        else:
            entry_keys = list(entry.keys())
            raise ValueError(f"Unknown dataset type with keys: {', '.join(entry_keys)}")

    if dataset_path.is_file():
        entries = load_jsonl(dataset_path, shuffle=shuffle)

        assert train_split + valid_split + test_split == 1.0, "Dataset splits must sum to 1.0"
        ix_train = int(len(entries) * train_split)
        ix_valid = int(len(entries) * (train_split + valid_split))
        ix_test = int(len(entries) * (train_split + valid_split + test_split))

        entries_split = {
            "train": entries[:ix_train],
            "valid": entries[ix_train:ix_valid],
            "test": entries[ix_valid:ix_test]
        }
    elif dataset_path.is_dir():
        entries_split = {}

        for name in ("train", "valid", "test"):
            dataset_file = dataset_path / f"{name}.jsonl"
            if dataset_file.exists():
                entries_split[name] = load_jsonl(dataset_file)
            else:
                raise FileNotFoundError(f"Dataset file not found: {dataset_file}")
    else:
        raise FileNotFoundError(f"Dataset file or directory not found: {dataset_path}")

    # Guess dataset type
    dataset_class = guess_type(entries_split["train"][0])

    # Load datasets
    datasets = []
    for name in ("train", "valid", "test"):
        dataset = dataset_class(entries_split[name], tokenizer, template=template, max_length=max_length)
        datasets.append(dataset)

        logger.info(f"Loaded {name} split for {dataset._type_id} dataset with {len(dataset)} entries")

    return datasets


from typing import Optional, Tuple

import mlx.core as mx
import mlx.nn as nn

from sillm.models.base import BaseModel
from sillm.models.args import ModelArgs
import sillm.models.mixtral as mixtral

class Attention(nn.Module):
    """
    Multi-head attention module.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        super().__init__()
        self.args = args

        self.dim: int = args.dim
        self.head_dim: int = args.head_dim
        self.n_heads: int = args.n_heads
        self.n_kv_heads: int = args.n_kv_heads
        self.clip_qkv: float = args.clip_qkv

        self.scale = self.args.head_dim ** -0.5

        self.wqkv = nn.Linear(args.dim, (self.n_kv_heads * 2 + self.n_heads) * self.head_dim, bias=False)
        self.wo = nn.Linear(args.dim, args.dim, bias=False)
        self.rope = nn.RoPE(args.head_dim,
                            traditional=args.rope_traditional,
                            base=args.rope_theta)
        
    def __call__(self,
                 x: mx.array,
                 mask: Optional[mx.array] = None,
                 cache: Optional[Tuple[mx.array, mx.array]] = None,
                 ) -> mx.array:
        B, L, _ = x.shape

        qkv = self.wqkv(x)
        qkv = mx.clip(qkv, a_min=-self.clip_qkv, a_max=self.clip_qkv)
        splits = [self.dim, self.dim + self.head_dim * self.n_kv_heads]
        queries, keys, values = mx.split(qkv, splits, axis=-1)

        queries = queries.reshape(B, L, self.n_heads, -1).transpose(0, 2, 1, 3)
        keys = keys.reshape(B, L, self.n_kv_heads, -1).transpose(0, 2, 1, 3)
        values = values.reshape(B, L, self.n_kv_heads, -1).transpose(
            0, 2, 1, 3
        )

        if cache is not None:
            key_cache, value_cache = cache
            queries = self.rope(queries, offset=key_cache.shape[2])
            keys = self.rope(keys, offset=key_cache.shape[2])
            keys = mx.concatenate([key_cache, keys], axis=2)
            values = mx.concatenate([value_cache, values], axis=2)
        else:
            queries = self.rope(queries)
            keys = self.rope(keys)

        output = mx.fast.scaled_dot_product_attention(
            queries, keys, values, scale=self.scale, mask=mask
        )
        output = output.transpose(0, 2, 1, 3).reshape(B, L, -1)

        return self.wo(output), (keys, values)

class TransformerBlock(mixtral.TransformerBlock):
    """
    Transformer block for DBRX models.
    """
    def __init__(self, args: ModelArgs):
        nn.Module.__init__(self)

        self.n_heads = args.n_heads
        self.dim = args.dim
        self.attention = Attention(args=args)
        self.feed_forward = mixtral.FeedForward(args=args)
        self.attention_norm = nn.LayerNorm(args.dim, bias=False)
        self.ffn_norm = nn.LayerNorm(args.dim, bias=False)
        self.args = args

class Model(mixtral.Model):
    """
    DBRX model wrapper.
    """
    def __init__(self, args: ModelArgs):
        """
        Args:
            args: Model arguments.
        """
        BaseModel.__init__(self, args)
        self.args = args

        self.n_layers = args.n_layers
        self.vocab_size = args.vocab_size
        self.router_aux_loss_coef = args.router_aux_loss_coef

        self.tok_embeddings = nn.Embedding(args.vocab_size, args.dim)
        self.layers = [TransformerBlock(args=args) for _ in range(args.n_layers)]
        self.norm = nn.LayerNorm(args.dim, bias=False)
        self.output = nn.Linear(args.dim, args.vocab_size, bias=False)


import pathlib
import json
import logging
import dataclasses

from sillm.utils.mapping import map_config

logger = logging.getLogger("sillm")

@dataclasses.dataclass
class ModelArgs:
    """
    Model arguments.
    """
    model_type: str
    dim: int
    n_layers: int
    head_dim: int
    n_heads: int
    n_kv_heads: int
    norm_eps: float = None
    hidden_dim: int = None
    vocab_size: int = -1
    rope_theta: float = 10000.0
    rope_traditional: bool = True
    max_position_embeddings: int = 0
    tie_word_embeddings: bool = False
    bos_token_id: int = None
    eos_token_id: int = None
    pad_token_id: int = None
    quantization: dict = None

    def __repr__(self):
        return json.dumps(dataclasses.asdict(self), indent=4)
    
    def log_config(self):
        for k, v in dataclasses.asdict(self).items():
            if isinstance(v, dict):
                for k2, v2 in v.items():
                    logger.debug(f"Config {k}.{k2}: {v2}")
            else:
                logger.debug(f"Config {k}: {v}")
    
    def fix_config(self, weights):
        """
        Fix config with shape information from weights.
        """
        if self.hidden_dim is None and "layers.0.feed_forward.w1.weight" in weights:
            self.hidden_dim = weights["layers.0.feed_forward.w1.weight"].shape[0]
        if self.vocab_size <= 0 and "output.weight" in weights:
            self.vocab_size = weights["output.weight"].shape[0]

    def save_config(self, config_path):
        """
        Save model config to JSON file.
        Args:
            config_path: Path to config file.
        """
        config = dataclasses.asdict(self)
        
        # Remove None values
        for k in list(config.keys()):
            if config[k] is None:
                del config[k]

        with open(config_path, "w") as f:
            f.write(json.dumps(config, indent=4))

    @staticmethod
    def load_config(config):
        ArgsClass = None
        if "model_type" in config:
            if config["model_type"] in ("llama", "mistral", "gemma"):
                ArgsClass = LlamaArgs
            elif config["model_type"] == "mixtral":
                ArgsClass = MixtralArgs
            elif config["model_type"] == "phi":
                ArgsClass = PhiArgs
            elif config["model_type"] == "qwen2":
                ArgsClass = Qwen2Args
            elif config["model_type"] == "starcoder2":
                ArgsClass = Starcoder2Args
            elif config["model_type"] == "dbrx":
                ArgsClass = DbrxArgs
            elif config["model_type"] == "cohere":
                ArgsClass = CohereArgs
            else:
                ArgsClass = LlamaArgs
        if ArgsClass is None:
            ArgsClass = LlamaArgs
            config["model_type"] = "llama"
            logger.warn(f"No model type specified - falling back to `llama` config")

        fields = ModelArgs.__annotations__
        fields.update(ArgsClass.__annotations__)
        config = {k:v for k, v in config.items() if k in fields}

        return ArgsClass(**config)
    
    @staticmethod
    def load_file(config_path):
        """
        Load model config from JSON file.
        Args:
            config_path: Path to config file.
        Returns:
            ModelArgs instance.
        """
        assert pathlib.Path(config_path).exists(), config_path

        with open(config_path, "r") as f:
            config = json.loads(f.read())
        config = map_config(config)

        return ModelArgs.load_config(config)
    
@dataclasses.dataclass
class LlamaArgs(ModelArgs):
    """
    Llama model arguments.
    """
    rope_scaling: dict = None

@dataclasses.dataclass
class MixtralArgs(ModelArgs):
    """
    Mixtral model arguments.
    """
    rope_theta: float = 1000000.0
    rope_scaling: dict = None
    router_aux_loss_coef: float = 0.001
    moe: dict = None

    def __post_init__(self):
        if self.moe is None:
            self.moe = {
                "num_experts": 8,
                "num_experts_per_tok": 2
            }

@dataclasses.dataclass
class PhiArgs(ModelArgs):
    """
    Phi model arguments.
    """
    rope_scaling: dict = None
    partial_rotary_factor: float = 0.4

@dataclasses.dataclass
class Qwen2Args(ModelArgs):
    """
    Starcoder2 model arguments.
    """
    rope_scaling: dict = None

@dataclasses.dataclass
class Starcoder2Args(ModelArgs):
    """
    Starcoder2 model arguments.
    """
    rope_scaling: dict = None
    tie_word_embeddings: bool = True

@dataclasses.dataclass
class DbrxArgs(ModelArgs):
    """
    DBRX model arguments.
    """
    clip_qkv: int = 8
    rope_theta: float = 500000.0
    router_aux_loss_coef: float = 0.05
    moe: dict = None

    def __post_init__(self):
        if self.moe is None:
            self.moe = {
                "num_experts": 16,
                "num_experts_per_tok": 4
            }

        if self.bos_token_id is None:
            self.bos_token_id = 100257
        if self.eos_token_id is None:
            self.eos_token_id = 100257

@dataclasses.dataclass
class CohereArgs(ModelArgs):
    """
    Cohere model arguments.
    """
    norm_bias: bool = False
    logit_scale: float = 0.0625
    use_qk_norm: bool = False


import matplotlib.pyplot as plt

class Plot:
    def __init__(self,
                 title="Training and Validation Losses",
                 ):
        """
        Initialize a plot for training and validation losses.
        Args:
            title: Title of the plot
        """
        self.title = title
        
        self.train_x = []
        self.train_y = []
        self.valid_x = []
        self.valid_y = []

    def add_train_loss(self, i, loss):
        """
        Add a training loss to the plot.
        Args:
            i: Step number
            loss: Training loss
        """
        self.train_x.append(i)
        self.train_y.append(loss)

    def add_valid_loss(self, i, val_loss):
        """
        Add a validation loss to the plot.
        Args:
            i: Step number
            val_loss: Validation loss
        """
        self.valid_x.append(i)
        self.valid_y.append(val_loss)

    def plot(self):
        """
        Plot the training and validation losses.
        """
        plt.figure(figsize=(15, 9))

        plt.xticks(ticks=self.valid_x)

        plt.plot(self.train_x, self.train_y, label='Training Loss')
        plt.plot(self.valid_x, self.valid_y, label='Validation Loss')

        plt.title(self.title)
        plt.xlabel('Steps')
        plt.ylabel('Loss')
        plt.legend()

    def save(self, fpath):
        """
        Save the plot to a file.
        Args:
            fpath: File path
        """
        self.plot()
        plt.savefig(fpath)

    def show(self):
        """
        Show the plot.
        """
        self.plot()
        plt.show()


import pathlib

from setuptools import setup, find_packages

pkg_dir = pathlib.Path(__file__).parent.resolve()
readme = (pkg_dir / "README.md").read_text(encoding="utf-8")
requirements = (pkg_dir / "requirements.txt").read_text(encoding="utf-8").splitlines()
requirements_server = (pkg_dir / "requirements-server.txt").read_text(encoding="utf-8").splitlines()

setup(
    name = "sillm-mlx",
    version = "0.1.2",
    description = "Running and training LLMs on Apple Silicon via MLX",
    long_description = readme,
    long_description_content_type = "text/markdown",
    readme="README.md",
    url = "https://github.com/armbues/SiLLM",
    install_requires=requirements,
    extras_require={
        "server": requirements_server
    },
    packages = find_packages(exclude="examples"),
    include_package_data=True,
    package_data = {
        'sillm': ['templates/*.jinja']
    },
    python_requires = ">=3.9"
)


import os
import argparse

import sillm
import sillm.utils as utils

if __name__ == "__main__":
    # Parse commandline arguments
    parser = argparse.ArgumentParser(description="A simple CLI for generating text with SiLLM.")
    parser.add_argument("model", type=str, help="The model directory or file")
    parser.add_argument("-d", "--chdir", default=None, type=str, help="Change working directory")
    parser.add_argument("-c", "--config", default=None, type=str, help="Load YAML configuration file for chat")
    parser.add_argument("-a", "--input_adapters", default=None, type=str, help="Load and merge LoRA adapter weights from .safetensors file")
    parser.add_argument("-s", "--seed", type=int, default=-1, help="Seed for randomization")
    parser.add_argument("-t", "--temperature", type=float, default=0.7, help="Sampling temperature")
    parser.add_argument("-f", "--flush", type=int, default=5, help="Flush output every n tokens")
    parser.add_argument("-m", "--max_tokens", type=int, default=1024, help="Max. number of tokens to generate")
    parser.add_argument("--template", type=str, default=None, help="Chat template (chatml, llama2, alpaca, etc.)")
    parser.add_argument("--system_prompt", type=str, default=None, help="System prompt for chat template")
    parser.add_argument("-q4", default=False, action="store_true", help="Quantize the model to 4 bits")
    parser.add_argument("-q8", default=False, action="store_true", help="Quantize the model to 8 bits")
    parser.add_argument("-v", "--verbose", default=1, action="count", help="Increase output verbosity")
    args = parser.parse_args()

    # Change working directory
    if args.chdir is not None:
        os.chdir(args.chdir)

    # Load YAML configuration file
    if args.config is not None:
        utils.load_yaml(args.config, args)
    
    # Initialize logging
    log_level = 40 - (10 * args.verbose) if args.verbose > 0 else 0
    logger = utils.init_logger(log_level)

    # Log commandline arguments
    if log_level <= 10:
        utils.log_arguments(args.__dict__)

    # Set random seed
    if args.seed >= 0:
        utils.seed(args.seed)

    # Load model
    model = sillm.load(args.model)

    if args.input_adapters is not None:
        # Convert model to trainable
        model = sillm.TrainableLoRA.from_model(model)

        lora_config = model.load_lora_config(args.input_adapters)

        # Initialize LoRA layers
        model.init_lora(**lora_config)

        # Load and merge adapter file
        model.load_adapters(args.input_adapters)
        model.merge_and_unload_lora()

    # Quantize model
    if args.q4 is True:
        model.quantize(bits=4)
    elif args.q8 is True:
        model.quantize(bits=8)

    generate_args = {
        "temperature": args.temperature,
        "max_tokens": args.max_tokens,
        "flush": args.flush
    }

    # Init conversation template
    template = sillm.init_template(model.tokenizer, model.args, args.template)
    conversation = sillm.Conversation(template, system_prompt=args.system_prompt)

    # Log memory usage
    utils.log_memory_usage()

    # Input loop
    while True:
        prompt = input("> ")

        if prompt.startswith('.'):
            break
        elif prompt == "":
            if conversation:
                conversation.clear()
            continue

        if conversation:
            prompt = conversation.add_user(prompt)
        
        logger.debug(f"Generating {args.max_tokens} tokens with temperature {args.temperature}")

        response = ""
        for s, metadata in model.generate(prompt, **generate_args):
            print(s, end="", flush=True)
            response += s
        print()

        if conversation:
            conversation.add_assistant(response)

        logger.debug(f"Evaluated {metadata['usage']['prompt_tokens']} prompt tokens in {metadata['timing']['eval_time']:.2f}s ({metadata['usage']['prompt_tokens'] / metadata['timing']['eval_time']:.2f} tok/sec)")
        logger.debug(f"Generated {metadata['usage']['completion_tokens']} tokens in {metadata['timing']['runtime']:.2f}s ({metadata['usage']['completion_tokens'] / metadata['timing']['runtime']:.2f} tok/sec)")


import numpy as np
import mlx.core as mx

import yaml

def seed(seed):
    """
    Seed random number generators.
    
    Args:
        seed: The seed value.
    """
    mx.random.seed(seed)
    np.random.seed(seed)

def load_yaml(fpath, args):
    with open(fpath, "r") as f:
        config = yaml.safe_load(f)

        for key, value in config.items():
            setattr(args, key, value)


import os
import pathlib
import time
import logging
import math
import re
import json

from functools import partial

import tqdm
import numpy as np

import mlx.core as mx
import mlx.nn as nn
import mlx.optimizers as optim
from mlx.utils import tree_flatten, tree_unflatten

from sillm.core.llm import LLM
from sillm.models.args import ModelArgs
from sillm.training.dataset import Dataset

logger = logging.getLogger("sillm")

########
# Based on mlx-examples:
# https://github.com/ml-explore/mlx-examples/blob/e74889d0fa0fb49d95bfdf6a1dcad907713eb50e/lora/models.py#L55
# https://github.com/ml-explore/mlx-examples/blob/854ad8747a9c703773adf8866602b114f68aa54a/llms/mlx_lm/tuner/lora.py#L7
########
class LoRALinear(nn.Module):
    """
    Linear layer with LoRA weights.
    """
    @staticmethod
    def from_linear(
        linear: nn.Linear,
        rank: int = 8,
        alpha: float = 16,
        dropout: float = 0.0,
        scale : float = 10.0
        ):
        """
        Convert linear layer to LoRA linear layer.
        Args:
            linear: Linear layer to convert.
            rank: Rank to use for LoRA.
            alpha: Alpha to use for LoRA.
            dropout: Dropout to use for LoRA.
            scale: Scale to use for LoRA.
        Returns:
            LoRA linear layer.
        """
        output_dims, input_dims = linear.weight.shape
        if isinstance(linear, nn.QuantizedLinear):
            input_dims *= 32 // linear.bits
        bias = "bias" in linear

        lora_lin = LoRALinear(input_dims, output_dims, rank, alpha, dropout, scale, bias)
        lora_lin.linear = linear
        lora_lin.name = linear.name

        return lora_lin

    def __init__(self,
                 input_dims: int,
                 output_dims: int,
                 rank: int = 8,
                 alpha: float = 16,
                 dropout: float = 0.0,
                 scale : float = 10.0,
                 bias: bool = False
                 ):
        """
        Args:
            input_dims: Input dimensions.
            output_dims: Output dimensions.
            rank: Rank to use for LoRA.
            alpha: Alpha to use for LoRA.
            dropout: Dropout to use for LoRA.
            scale: Scale to use for LoRA.
            bias: Whether to use bias.
        """
        super().__init__()

        # Initialize linear layer weights
        self.linear = nn.Linear(input_dims, output_dims, bias=bias)

        # Initialize LoRA dropout
        self.lora_dropout = nn.Dropout(p=dropout)

        # Initialize LoRA weights
        self.scale = scale * (alpha / rank)
        bound = 1 / math.sqrt(input_dims)
        input_shape = (input_dims, rank)
        output_shape = (rank, output_dims)
        self.lora_a = mx.random.uniform(low=-bound, high=bound, shape=input_shape)
        self.lora_b = mx.zeros(shape=output_shape)

    @property
    def lora_size(self):
        """
        Returns:
            Number of LoRA parameters.
        """
        return self.lora_a.size + self.lora_b.size

    def __call__(self, x):
        """
        Args:
            x: Input tensor.
        Returns:
            Output tensor.
        """
        # Determine dtype
        dtype = self.linear.weight.dtype
        if isinstance(self.linear, nn.QuantizedLinear):
            dtype = self.linear.scales.dtype

        # Apply linear layer and LoRA
        y = self.linear(x.astype(dtype))
        z = (self.lora_dropout(x) @ self.lora_a) @ self.lora_b

        return y + self.scale * z
    
    def merge(self):
        """
        Merge LoRA weights into linear weights.
        Returns:
            Linear layer with merged weights.
        """
        linear = self.linear
        weight = linear.weight
        dtype = linear.weight.dtype

        quantized = isinstance(linear, nn.QuantizedLinear)
        if quantized:
            dtype = mx.float16
            group_size = linear.group_size
            bits = linear.bits
            weight = mx.dequantize(weight, linear.scales, linear.biases, group_size, bits)

        # Merge LoRA weights into linear weights
        update = (self.lora_a @ self.lora_b).transpose()
        weight = (weight + (self.scale * update)).astype(dtype)

        if quantized:
            output_dims, input_dims = weight.shape
            bias = "bias" in linear
            linear = nn.Linear(input_dims, output_dims, bias=bias)

            return nn.QuantizedLinear.from_linear(linear, group_size, bits)
        else:
            linear.weight = weight

            return linear

class TrainableLoRA(LLM):
    """
    Trainable LoRA model wrapper.
    """
    @staticmethod
    def from_model(llm: LLM):
        """
        Convert LLM to trainable LLM.
        Args:
            llm: LLM to convert.
        Returns:
            Trainable LLM.
        """
        model = TrainableLoRA(llm.model, llm.tokenizer, llm.args)
        model._quantization = llm._quantization
        model._mapping = llm._mapping

        return model
    
    def __init__(self,
                 model,
                 tokenizer,
                 args: ModelArgs
                 ):
        """
        Args:
            tokenizer: Tokenizer instance.
            args: Model arguments.
        """
        self.model = model
        self.tokenizer = tokenizer
        self.args = args

        self._lora = None
        self._lora_modules = []

    def init_lora(self,
                  num_layers: int = -1,
                  target_modules: str = "query_value",
                  rank: int = 8,
                  alpha: float = 16,
                  dropout: float = 0.0,
                  scale : float = 10.0
                  ):
        """
        Initialize LoRA for model.
        Args:
            num_layers: Number of layers to apply LoRA to.
            target_modules: Modules to apply LoRA to.
            rank: Rank to use for LoRA.
            alpha: Alpha to use for LoRA.
            dropout: Dropout to use for LoRA.
            scale: Scale to use for LoRA.
        """
        if self._lora is None:
            self.model.freeze()

            if num_layers < 0:
                num_layers = len(self.model.layers)

            self._lora = {
                "num_layers": num_layers,
                "target_modules": target_modules,
                "rank": rank,
                "alpha": alpha,
                "dropout": dropout,
                "scale": scale
            }

            if target_modules == "all_linear":
                self._lora_modules = [
                    (key, LoRALinear.from_linear(module, rank=rank, alpha=alpha, dropout=dropout, scale=scale))
                    for key, module in self.model.named_modules()
                    if isinstance(module, nn.Linear) or isinstance(module, nn.QuantizedLinear)
                ]
            elif target_modules == "query_value":
                self._lora_modules = [
                    (key, LoRALinear.from_linear(module, rank=rank, alpha=alpha, dropout=dropout, scale=scale))
                    for key, module in self.model.named_modules()
                    if re.search(r"\.attention\.(wq|wv)$", key)
                ]
            if len(self._lora_modules) == 0:
                logger.error(f"No target modules found for LoRA: {target_modules}")
            self.model.update_modules(tree_unflatten(self._lora_modules))

            # Enable training mode
            self.model.train(mode=True)

            logger.info(f"Initialized LoRA with rank {rank} for {num_layers} layers")
            logger.debug(f"LoRA target modules: {target_modules}")
            logger.debug(f"LoRA parameters: Alpha {alpha}, Dropout {dropout}, Scale {scale}")

            trainable_params = 0
            for _, module in self._lora_modules:
                trainable_params += module.lora_size
            logger.debug(f"LoRA trainable parameters: {trainable_params/ 10**6:.2f}M")

    def merge_and_unload_lora(self):
        """
        Merge LoRA layers back into model.
        """
        if self._lora is not None:
            merged_modules = [
                (key, module.merge())
                for key, module in self._lora_modules
            ]
            self.model.update_modules(tree_unflatten(merged_modules))

            logger.info(f"Merged LoRA layers back into model")

        self._lora = None
        self._lora_modules = []

        # Disable training mode
        self.model.train(mode=False)

    def save_lora_config(self,
                         config_path: str
                         ):
        """
        Save LoRA configuration.
        Args:
            config_path: Folder to save LoRA configuration to.
        """
        assert self._lora is not None

        config_path = pathlib.Path(config_path) / "lora.json"

        with open(config_path, "w") as f:
            f.write(json.dumps(self._lora))

    def load_lora_config(self,
                         config_path: str
                         ):
        """
        Load LoRA configuration.
        Args:
            config_path: Path to load LoRA configuration from.
        """
        config_path = pathlib.Path(config_path)
        if config_path.is_file():
            if config_path.suffix == ".safetensors":
                config_path = config_path.parent
        if config_path.is_dir():
            config_path = config_path / "lora.json"

        if config_path.exists():
            with open(config_path, "r") as f:
                logger.info(f"Loaded LoRA configuration from {config_path}")

                return json.loads(f.read())
            
        return {}

    def save_adapters(self,
                      adapter_path: str,
                      ):
        """
        Save adapter weights.
        Args:
            adapter_path: Path to save adapter weights to.
        """
        assert self._lora is not None

        state = dict(tree_flatten(self.model.trainable_parameters()))
        metadata = {
            "format": "mlx"
        }
        mx.save_safetensors(adapter_path, state, metadata=metadata)

    def save_checkpoint(self,
                        checkpoint_path: str,
                        steps: int = -1
                        ):
        """
        Save model checkpoint.
        Args:
            checkpoint_path: Directory to save checkpoints to.
            steps: Number of steps.
        """
        checkpoint_path = pathlib.Path(checkpoint_path)
        if steps >= 0:
            adapter_path = checkpoint_path / f"ckpt-{steps}.safetensors"
        else:
            adapter_path = checkpoint_path / f"ckpt-final.safetensors"

        self.save_adapters(str(adapter_path))
        
        return str(adapter_path)

    def load_adapters(self,
                      adapter_path: str
                      ):
        """
        Load adapter weights.
        Args:
            adapter_path: Path to adapter weights.
        """
        assert pathlib.Path(adapter_path).exists(), adapter_path

        if self._lora is not None:
            self.model.load_weights(adapter_path, strict=False)
        
            logger.info(f"Loaded adapter weights from {adapter_path}")
        else:
            logger.error(f"LoRA not initialized, cannot load adapter weights")

    ########
    # Based on mlx-examples:
    # https://github.com/ml-explore/mlx-examples/blob/e74889d0fa0fb49d95bfdf6a1dcad907713eb50e/lora/lora.py#L198
    ########
    def evaluate(self,
                 dataset: Dataset,
                 batch_size: int,
                 num_batches: int
                 ):
        """
        Evaluate model on dataset.
        Args:
            dataset: Dataset to evaluate on.
            batch_size: Batch size.
            num_batches: Number of batches to evaluate.
        Returns:
            Average loss.
        """
        all_losses = []
        num_tokens = 0
        for _, batch in zip(
            range(num_batches),
            dataset.iterate_batches(batch_size),
        ):
            losses, _, toks = self.loss(*batch)
            all_losses.append((losses * toks).item())
            num_tokens += toks.item()

        return np.sum(all_losses) / num_tokens
    
    def loss(self, *args, **kwargs):
        """
        Default loss function from model.
        """
        return self.model.loss(*args, **kwargs)

    ########
    # Based on mlx-examples:
    # https://github.com/ml-explore/mlx-examples/blob/e74889d0fa0fb49d95bfdf6a1dcad907713eb50e/lora/lora.py#L212
    ########
    def train(self, 
              dataset_training: Dataset,
              dataset_validation: Dataset,
              batch_size: int = 4,
              learning_rate: float = 1e-5,
              compiled_step: bool = True,
              grad_checkpoint: bool = False,
              epochs: int = 1,
              iterations: int = 0,
              report_steps: int = 10,
              report_callback: callable = None,
              eval_steps: int = 100,
              eval_callback: callable = None,
              validation_samples: int = 40
              ):
        """
        Train model.
        Args:
            dataset_training: Training dataset.
            dataset_validation: Validation dataset.
            batch_size: Batch size.
            learning_rate: Learning rate.
            epochs: Number of epochs.
            iterations: Number of iterations.
            report_steps: Report every `report_steps` iterations.
            eval_steps: Evaluate every `eval_steps` iterations.
            eval_callback: Callback after eval.
            validation_samples: Number of validation samples.
            debug: Whether to enable debug mode.
        """
        # Calculate number of iterations
        if iterations == 0:
            iterations = len(dataset_training) // batch_size
        
        # Calculate number of validation batches
        validation_batches = validation_samples // batch_size
        
        logger.info(f"Training the model for {epochs} epochs of {iterations} batch iterations with batch size {batch_size}")
        logger.debug(f"Training learning rate: {learning_rate}")

        optimizer = optim.Adam(learning_rate=learning_rate)

        if grad_checkpoint:
            logger.info(f"Enabled gradient checkpointing")

            if not compiled_step:
                logger.warning(f"Gradient checkpointing requires compiled step function")
                compiled_step = True
            
            for layer in self.model.layers:
                layer.forward = nn.utils.checkpoint(layer, layer.forward)

        # Create value and gradient function for loss
        loss_value_and_grad = nn.value_and_grad(self.model, self.loss)

        if compiled_step:
            state = [self.model.state, optimizer.state]

            # Step function for forward and backward pass
            @partial(mx.compile, inputs=state, outputs=state)
            def step(batch):
                (loss_value, reward, num_tokens), grad = loss_value_and_grad(*batch)
                optimizer.update(self.model, grad)

                return loss_value, reward, num_tokens

        # Get system memory
        system_memory = os.sysconf("SC_PAGE_SIZE") * os.sysconf("SC_PHYS_PAGES")

        losses = []
        rewards = None
        intv_tokens = 0

        # Main training loop
        start = time.perf_counter()
        pbar_epochs = tqdm.tqdm(range(epochs), desc="Epoch")
        for epoch in pbar_epochs:
            pbar_iterations = tqdm.tqdm(range(iterations), desc="Iter.", leave=False)
            for iter in pbar_iterations:
                n = epoch * iterations + iter
                batch = next(dataset_training.iterate_batches(batch_size, train=True))

                if compiled_step:
                    loss_value, reward, num_tokens = step(batch)
                else:
                    (loss_value, reward, num_tokens), grad = loss_value_and_grad(*batch)
                    optimizer.update(self.model, grad)

                mx.eval(loss_value, reward, num_tokens)

                # Record loss and number of tokens
                losses.append(loss_value.item())
                intv_tokens += num_tokens.item()

                # Record rewards
                if reward is not None:
                    if rewards is None:
                        rewards = reward
                    else:
                        rewards = np.vstack([rewards, reward])

                # Get peak memory usage
                peak_memory = mx.metal.get_peak_memory()
                memory_usage = peak_memory / system_memory
                if memory_usage > 0.9:
                    pbar_epochs.write(f"HIGH MEMORY USAGE: {(peak_memory // (1024 ** 2)):,} MB ({memory_usage:.2%} of system memory)")

                # Report training loss if needed
                if (n + 1) % report_steps == 0:
                    train_loss = np.mean(losses)
                    stop = time.perf_counter()

                    pbar_epochs.write(f"#{n + 1}:\tTraining loss    {train_loss:.3f}\t{float(intv_tokens) / (stop - start):.3f} tok/sec")
                    if rewards is not None:
                        pbar_epochs.write(f"#{n + 1}:\tTraining reward  {str(np.mean(rewards, axis=0))}")
                        rewards = None
                    pbar_epochs.refresh()

                    if report_callback is not None:
                        report_callback(n + 1, train_loss)
                    
                    losses = []
                    intv_tokens = 0
                    start = time.perf_counter()

                # Report validation loss if needed
                if n == 0 or (n + 1) % eval_steps == 0:
                    stop = time.perf_counter()
                    val_loss = self.evaluate(dataset_validation, batch_size, validation_batches)
                    start = time.perf_counter()
                    pbar_epochs.write(f"#{n + 1}:\tValidation loss  {val_loss:.3f}\t{(start - stop):.3f} sec")
                    pbar_epochs.write(f"#{n + 1}:\tPeak memory      {(peak_memory // (1024 ** 2)):,} MB ({memory_usage:.2%} of system memory)")

                    # Eval callback
                    if eval_callback is not None:
                        msg = eval_callback(n + 1, val_loss)
                        if msg:
                            pbar_epochs.write(f"#{n + 1}:\t" + msg)

                    start = time.perf_counter()


import argparse
import time
import uuid

import fastapi
import uvicorn
import pydantic

import sillm
import sillm.utils as utils

app = fastapi.FastAPI()

@app.get("/")
def index():
    return {
        "message": "Welcome to the SiLLM API!"
    }

@app.get("/v1/models")
def list_models():
    return {
        "object": "list",
        "data": [app.model.description()]
    }

class ChatCompletion(pydantic.BaseModel):
    messages: list
    model: str
    logprobs: bool | None = False
    max_tokens: int | None = 4096
    temperature: float | None = 0.7

@app.post("/v1/chat/completions")
def chat_completions(completion: ChatCompletion):
    result = {
        "id": str(uuid.uuid4()),
        "object": "chat.completion",
        "created": int(time.time()),
        "model": app.model.id,
        "system_fingerprint": app.fingerprint
    }

    prompt = app.template.apply_chat_template(completion.messages, add_generation_prompt=True)
    
    generate_args = {
        "max_tokens": completion.max_tokens,
        "temperature": completion.temperature
    }

    content = ""
    for s, metadata in app.model.generate(prompt, **generate_args):
        content += s

    result["choices"] = [{
        "index": 0,
        "message": {
            "role": "assistant",
            "content": content
        },
        "logprobs": None,
        "finish_reason": metadata["finish_reason"]
    }]
    result["usage"] = metadata["usage"]

    return result

if __name__ == "__main__":
    # Parse commandline arguments
    parser = argparse.ArgumentParser(description="")
    parser.add_argument("model", type=str, help="The directory that contains the models")
    parser.add_argument("--host", type=str, default="0.0.0.0", help="Port to listen on")
    parser.add_argument("--port", type=int, default=8000, help="Port to listen on")
    parser.add_argument("--template", type=str, default=None, help="Chat template (chatml, llama2, alpaca, etc.)")
    parser.add_argument("-v", "--verbose", default=1, action="count", help="Increase output verbosity")
    args = parser.parse_args()

    # Initialize logging
    log_level = 40 - (10 * args.verbose) if args.verbose > 0 else 0
    logger = utils.init_logger(log_level)
    app.logger = logger
    
    # Load model
    model = sillm.load(args.model)
    app.model = model

    # Initialize system fingerprint
    app.fingerprint = "sillm_" + str(uuid.uuid4())

    # Init conversation template
    app.template = sillm.init_template(model.tokenizer, model.args, args.template)

    uvicorn.run(app, host=args.host, port=args.port)